import json
import os
import shutil
import warnings
from abc import ABC, abstractmethod
from os.path import join
from typing import Any, Callable, Dict, Optional, Tuple, TypeVar, Union

import numpy as np
from copy import deepcopy
import pandas as pd
import torch
from PIL import Image, UnidentifiedImageError
from sympy import false
from torch.utils.data import DataLoader, Dataset
from torchtext import datasets as textds
from torchvision import datasets
from torchvision.transforms import Resize, ToTensor
from tqdm import tqdm


from gdeep.utility.custom_types import Tensor

T = TypeVar("T")


class ImageClassificationFromFiles(Dataset[Any]):
    """This class is useful to build a dataset
    directly from image files

    Args:
        img_folder (string):
            The path to the folder where the training
            images are located
        labels_file (string):
            The path and file name of the labels.
            It shall be a ``.csv`` file with two columns.
            The first columns contains the name of the
            image and the second one contains the
            label value
        transform (AbstractPreprocessing):
            the instance of the class of preprocessing.
            It inherits from ``AbstractPreprocessing``
        target_transform (AbstractPreprocessing):
            the instance of the class of preprocessing.
            It inherits from ``AbstractPreprocessing``
    """

    img_folder: str
    labels_file: str
    img_labels: pd.DataFrame

    def __init__(
        self,
        img_folder: str = ".",
        labels_file: str = "labels.csv",
    ) -> None:

        self.img_folder = img_folder
        self.img_labels = pd.read_csv(labels_file)

    def __len__(self) -> int:
        return len(self.img_labels)

    def __getitem__(self, idx: int) -> Tuple[Any, Union[str, int]]:
        image = self._get_image(idx)
        image_out = deepcopy(image)
        label = self.img_labels.iloc[idx, 1]
        image.close()
        return image_out, label  # type: ignore

    def _get_image(self, idx: int) -> Any:
        """this method gets the i-th image in the labels.csv
        file.
        """
        img_path: str = os.path.join(self.img_folder, self.img_labels.iloc[idx, 0])  # type: ignore
        try:
            image = Image.open(img_path)
        except UnidentifiedImageError:
            warnings.warn(f"The image {img_path} canot be loaded. Skipping it.")
            return None
        return image



from typing import Any, Optional, Tuple, TypeVar, Union

from torchtext.data import to_map_style_dataset
import torch
from torch.utils.data import DataLoader, IterDataPipe, Dataset, MapDataPipe
from torchvision import datasets
from torchtext import datasets as textdatasets
from .tori import ToriDataset
from gdeep.utility import DEFAULT_DOWNLOAD_DIR
from ..dataset_factory import DatasetFactory
from torch.utils.data import IterableDataset

from gdeep.utility.custom_types import Tensor

T = TypeVar("T")


class TorchvisionDatasetBuilder(object):
    """Builder class for the torchvision dataset
    and all its variations"""

    def __init__(self, dataset_name: str):
        self.dataset_name = dataset_name

    def __call__(self, **kwargs):  # type: ignore
        return datasets.__getattribute__(self.dataset_name)(  # type: ignore
            root=DEFAULT_DOWNLOAD_DIR, **kwargs  # type: ignore
        )


class TorchtextDatasetBuilder(object):
    """Builder class for the torchtext dataset
    and all its variations"""

    def __init__(self, dataset_name: str):
        self.dataset_name = dataset_name

    def __call__(self, **kwargs):  # type: ignore
        return textdatasets.__getattribute__(self.dataset_name)(  # type: ignore
            root=DEFAULT_DOWNLOAD_DIR, **kwargs  # type: ignore
        )


class ToriDatasetBuilder(object):
    """Builder class for the torus dataset
    and all its variations"""

    def __init__(self, name: str) -> None:
        self.name = name

    def __call__(self, **kwargs):  # type: ignore
        return ToriDataset(name=self.name, **kwargs)


def get_dataset(key: str, **kwargs) -> Tuple[Dataset[Any]]:
    """Get a dataset from the factory

    Args:
        key :
            The name of the dataset, correspondiong
            to the key in the list of builders
        **kwargs:
            The keyword arguments to pass to the dataset builder

    Returns:
        torch.utils.data.Dataset:
            The dataset
    """
    factory = DatasetFactory()

    factory.register_builder("DoubleTori", ToriDatasetBuilder("DoubleTori"))
    factory.register_builder("Blobs", ToriDatasetBuilder("Blobs"))
    factory.register_builder("EntangledTori", ToriDatasetBuilder("EntangledTori"))

    for dataset_name in datasets.__all__:
        factory.register_builder(dataset_name, TorchvisionDatasetBuilder(dataset_name))

    for dataset_name in textdatasets.__all__:
        factory.register_builder(dataset_name, TorchtextDatasetBuilder(dataset_name))

    return factory.build(key, **kwargs)  # type: ignore


class DatasetBuilder:
    """Class to obtain Datasets from the classical
    datasets available on pytorch. Also the torus dataset
    and all its variations can be found here

    Args:
        name:
            check the available datasets at
            https://pytorch.org/vision/stable/datasets.html
            and https://pytorch.org/text/stable/datasets.html
        convert_to_map_dataset:
            whether to conver to the MapDataset or to keep
            IterableDataset

    """

    train_ds: Dataset[Any]
    valid_ds: Optional[Dataset[Any]]
    test_ds: Optional[Dataset[Any]]

    def __init__(
        self, name: str = "MNIST", convert_to_map_dataset: bool = False
    ) -> None:
        self.convert_to_map_dataset = convert_to_map_dataset
        self.name = name

    def build(
        self, **kwargs
    ) -> Tuple[Dataset[Any], Optional[Dataset[Any]], Optional[Dataset[Any]]]:
        """Method that returns the dataset.

        Args:
            kwargs:
                the arguments to pass to the dataset builder.
                For example, you may want to use the options
                ``split=("train","dev")`` or ``split=("train","test")``
        """
        dataset_tuple: Tuple[Dataset[Any]] = get_dataset(self.name, **kwargs)  # type: ignore

        if len(dataset_tuple) == 1:
            self.train_ds = dataset_tuple[0]
            self.valid_ds = None
            self.test_ds = None
        elif len(dataset_tuple) == 2:
            self.train_ds = dataset_tuple[0]
            self.valid_ds = dataset_tuple[1]  # type: ignore
            self.test_ds = None
        elif len(dataset_tuple) == 3:
            self.train_ds = dataset_tuple[0]
            self.valid_ds = dataset_tuple[1]  # type: ignore
            self.test_ds = dataset_tuple[2]  # type: ignore
        else:
            self.train_ds = dataset_tuple  # type: ignore
            self.valid_ds = None
            self.test_ds = None

        if self.convert_to_map_dataset:
            self.train_ds, self.valid_ds, self.test_ds = self._convert(
                self.train_ds, self.valid_ds, self.test_ds
            )
        return self.train_ds, self.valid_ds, self.test_ds

    def _convert(
        self,
        training_data: Dataset[Any],
        validation_data: Optional[Dataset[Any]] = None,
        test_data: Optional[Dataset[Any]] = None,
    ) -> Tuple[Dataset[Any], Optional[Dataset[Any]], Optional[Dataset[Any]]]:
        """This private method converts and IterableDataset
        to a MapDataset"""
        if isinstance(training_data, IterableDataset):
            training_data = self._to_map_style_dataset(training_data)
        if validation_data is not None:
            if isinstance(validation_data, IterableDataset):
                validation_data = self._to_map_style_dataset(validation_data)
        if test_data is not None:
            if isinstance(test_data, IterableDataset):
                test_data = self._to_map_style_dataset(test_data)
        return training_data, validation_data, test_data

    @staticmethod
    def _to_map_style_dataset(
        dataset: Union[Dataset, IterDataPipe]
    ) -> Union[Dataset, MapDataPipe]:
        """our version of to_map_style_dataset, with the
        possibility to handle DataPipe and Dataset as well
        """
        # if isinstance(dataset, IterDataPipe):
        #    return dataset.to_map_datapipe()
        return to_map_style_dataset(dataset)



import json
import os
import shutil
from os.path import join
from typing import Any, Dict, Tuple, TypeVar, Union, List

import numpy as np
import torch
from torch.utils.data import DataLoader

from .dataset_form_array import FromArray
from .dataset_cloud import DatasetCloud
from .base_dataloaders import DataLoaderBuilder
from .base_dataloaders import AbstractDataLoaderBuilder

from gdeep.utility.custom_types import Tensor

T = TypeVar("T")


class DlBuilderFromDataCloud(AbstractDataLoaderBuilder):
    """Class that loads data from Google Cloud Storage

    This class is useful to build dataloaders from a dataset stored in
    the GDeep Dataset Cloud on Google Cloud Storage.

    The constructor takes the name of a dataset as a string, and a string
    for the download directory. The constructor will download the dataset
    to the download directory. The dataset is downloaded in the version
    used by Datasets Cloud, which may be different from the version
    used by the dataset's original developers.

    Args:
        dataset_name (str):
            The name of the dataset.
        download_dir (str):
            The directory where the dataset will be downloaded.
        use_public_access (bool):
            Whether to use public access. If you want
            to use the Google Cloud Storage API, you must set this to True.
            Please make sure you have the appropriate credentials.
        path_to_credentials (str):
            Path to the credentials file.
            Only used if public_access is False and credentials are not
            provided. Defaults to None.


    Returns:
        torch.utils.data.DataLoader: The dataloader for the dataset.

    Raises:
        ValueError:
            If the dataset_name is not a valid dataset that exists
            in Datasets Cloud.
        ValueError:
            If the download_directory is not a valid directory.
    """

    def __init__(
        self,
        dataset_name: str,
        download_directory: str,
        use_public_access: bool = True,
        path_to_credentials: Union[None, str] = None,
    ):
        self.dataset_name = dataset_name
        self.download_directory = download_directory

        # Download the dataset if it does not exist
        self.download_directory

        self._download_dataset(
            use_public_access=use_public_access, path_to_credentials=path_to_credentials
        )

        self.dl_builder = None

        # Load the metadata of the dataset
        with open(
            join(self.download_directory, self.dataset_name, "metadata.json")
        ) as f:
            self.dataset_metadata = json.load(f)

        # Load the data and labels of the dataset
        if self.dataset_metadata["data_type"] == "tabular":
            if self.dataset_metadata["data_format"] == "pytorch_tensor":
                data = torch.load(
                    join(self.download_directory, self.dataset_name, "data.pt")
                )
                labels = torch.load(
                    join(self.download_directory, self.dataset_name, "labels.pt")
                )

                self.dl_builder = DataLoaderBuilder(
                    [
                        FromArray(data, labels),
                    ]
                )
            elif self.dataset_metadata["data_format"] == "numpy_array":
                data = np.load(
                    join(self.download_directory, self.dataset_name, "data.npy")
                )
                labels = np.load(
                    join(self.download_directory, self.dataset_name, "labels.npy")
                )
                self.dl_builder = DataLoaderBuilder(
                    [
                        FromArray(data, labels),  # type: ignore
                    ]
                )
            else:
                raise ValueError(
                    f"Data format {self.dataset_metadata['data_format']}"
                    + "is not yet supported."
                )
        else:
            raise ValueError(
                f"Dataset type {self.dataset_metadata['data_type']} is not yet supported."
            )

    def _download_dataset(
        self,
        path_to_credentials: Union[None, str] = None,
        use_public_access: bool = True,
    ) -> None:
        """Only download if the download directory does not exist already
        and if download directory contains at least three files (metadata,
        data, labels).

        Args:
            path_to_credentials (str):
                Path to the credentials file.
            use_public_access (bool):
                Whether to use public access. If you want
                to use the Google Cloud Storage API, you must set this to True.

        Returns:
            None
        """
        if (
            not os.path.isdir(join(self.download_directory, self.dataset_name))
            or len(os.listdir(join(self.download_directory, self.dataset_name))) < 3
        ):
            # Delete the download directory if it exists but does not contain
            # the wanted number of files
            if (
                os.path.isdir(join(self.download_directory, self.dataset_name))
                and len(os.listdir(join(self.download_directory, self.dataset_name)))
                < 3
            ):  # type: ignore
                print(
                    "Deleting the download directory because it does "
                    + "not contain the dataset"
                )
                shutil.rmtree(self.download_directory, ignore_errors=True)

            print(
                "Downloading dataset {} to {}".format(
                    self.dataset_name, self.download_directory
                )
            )
            dataset_cloud = DatasetCloud(
                self.dataset_name,
                download_directory=self.download_directory,
                path_to_credentials=path_to_credentials,
                use_public_access=use_public_access,
            )
            dataset_cloud.download()
            del dataset_cloud
        else:
            print("Dataset '%s' already downloaded" % self.dataset_name)

    def get_metadata(self) -> Dict[str, Any]:
        """Returns the metadata of the dataset.

        Returns:
            Dict[str, Any]:
                The metadata of the dataset.
        """
        return self.dataset_metadata

    def build(
        self, tuple_of_kwargs: List[Dict[str, Any]]
    ) -> Tuple[DataLoader, DataLoader, DataLoader]:
        """Builds the dataloaders for the dataset.

        Args:
            **tuple_of_kwargs: Arguments for the dataloader builder.

        Returns:
            Tuple[DataLoader, DataLoader, DataLoader]:
                The dataloaders for the dataset (train, validation, test).
        """
        return self.dl_builder.build(tuple_of_kwargs)  # type: ignore



from typing import Optional

import torch
from torch.nn import Module

from ..persformer_config import PersformerConfig

# Type aliases
from gdeep.utility.custom_types import Tensor


class MaxPoolingLayer(Module):
    """Implementation of the max pooling layer"""

    config: PersformerConfig

    def __init__(self, config: PersformerConfig):
        super().__init__()
        self.config = config

    def forward(
        self, input_batch: Tensor, attention_mask: Optional[Tensor] = None
    ) -> Tensor:
        """
        Forward pass of the model.

        Args:
            input_batch:
                The input batch. Of shape (batch_size, sequence_length, hidden_size)

        Returns:
            The logits of the model. Of shape (batch_size, sequence_length, 1)
        """
        # Initialize the output tensor
        if attention_mask is not None:
            output = input_batch * attention_mask.unsqueeze(2)
        else:
            output = input_batch
        # Apply the max pooling layer
        output = output.max(dim=-2)[0]
        return output



import os
from os import remove
from os.path import join, exists
from typing import List, Tuple, Union, Set

import json
import wget  # type: ignore

from ._data_cloud import _DataCloud  # type: ignore
from gdeep.utility.constants import DEFAULT_DOWNLOAD_DIR, DATASET_BUCKET_NAME


class DatasetCloud:
    """DatasetCloud class to handle the download and upload
    of datasets to the DataCloud.
    If the download_directory does not exist, it will be created and
    if a folder with the same name as the dataset exists in the
    download directory, it will not be downloaded again.
    If a folder with the same name as the dataset does not exists
    locally, it will be created when downloading the dataset.

    Args:
        dataset_name (str):
            Name of the dataset to be downloaded or uploaded.
        bucket_name (str, optional):
            Name of the bucket in the DataCloud.
            Defaults to DATASET_BUCKET_NAME.
        download_directory (Union[None, str], optional):
            Directory where the
            dataset will be downloaded to. Defaults to DEFAULT_DOWNLOAD_DIR.
        use_public_access (bool, optional):
            If True, the dataset will
            downloaded via public url. Defaults to False.
        path_credentials (Union[None, str], optional):
            Path to the credentials file.
            Only used if public_access is False and credentials are not
            provided. Defaults to None.
        make_public (bool, optional):
            If True, the dataset will be made public

        Raises:
            ValueError:
                Dataset does not exits in cloud.

        Returns:
            None
    """

    def __init__(
        self,
        dataset_name: str,
        bucket_name: str = DATASET_BUCKET_NAME,
        download_directory: Union[None, str] = None,
        use_public_access: bool = True,
        path_to_credentials: Union[None, str] = None,
        make_public: bool = True,
    ) -> None:
        # Non-public datasets start with "private_"
        if make_public or use_public_access or dataset_name.startswith("private_"):
            self.name = dataset_name
        else:
            self.name = "private_" + dataset_name
        self.path_metadata = None
        self.use_public_access = use_public_access
        if download_directory is None:
            # If download_directory is None, the dataset will be downloaded
            # to the default directory.
            self.download_directory = DEFAULT_DOWNLOAD_DIR
        else:
            self.download_directory = download_directory

        self._data_cloud = _DataCloud(
            bucket_name=bucket_name,
            download_directory=self.download_directory,
            use_public_access=use_public_access,
            path_to_credentials=path_to_credentials,
        )
        if use_public_access:
            self.public_url = "https://storage.googleapis.com/" + bucket_name + "/"
        self.make_public = make_public

    def __del__(self) -> None:
        """This function deletes the metadata file if it exists.

        Returns:
            None
        """
        if self.path_metadata != None:
            remove(self.path_metadata)  # type: ignore
        return None

    def download(self) -> None:
        """Download a dataset from the DataCloud. If the dataset does not
        exist in the cloud, an exception will be raised. If the dataset
        exists locally in the download directory, the dataset will not be
        downloaded again.

        Raises:
            ValueError:
                Dataset does not exits in cloud.
            ValueError:
                Dataset exists locally but checksums do not match.
        """
        if self.use_public_access:
            self._download_using_url()
        else:
            self._download_using_api()

    def _download_using_api(self) -> None:
        """Downloads the dataset using the DataCloud API.
        If the dataset does not exist in the bucket, an exception will
        be raised. If the dataset exists locally in the download directory,
        the dataset will not be downloaded again.

        Raises:
            ValueError:
                Dataset does not exits in cloud.

        Returns:
            None
        """
        self._check_public_access()
        # List of existing datasets in the cloud.
        existing_datasets: Set[str] = set(
            [
                blob.name.split("/")[0]  # type: ignore
                for blob in self._data_cloud.bucket.list_blobs()  # type: ignore  # type: ignore
                if blob.name != "giotto-deep-big.png"
            ]
        )  # type: ignore
        if self.name not in existing_datasets:
            raise ValueError(
                "Dataset {} does not exist in the cloud.".format(self.name)
                + "Available datasets are: {existing_datasets}."
            )
        if not self._does_dataset_exist_locally():
            self._create_dataset_folder()
        self._data_cloud.download_folder(self.name + "/")

    def _does_dataset_exist_locally(self) -> bool:
        """Check if the dataset exists locally.

        Returns:
            bool: True if the dataset exists locally, False otherwise.
        """
        return exists(join(self.download_directory, self.name))

    def _create_dataset_folder(self) -> None:
        """Creates a folder with the dataset name in the download directory.

        Returns:
            None
        """
        if not exists(join(self.download_directory, self.name)):
            os.makedirs(join(self.download_directory, self.name))

    def _download_using_url(self) -> None:
        """Download the dataset using the public url.
        If the dataset does not exist in the bucket, an exception will
        be raised. If the dataset exists locally in the download directory,
        the dataset will not be downloaded again.

        Raises:
            ValueError:
                Dataset does not exits in cloud.

        Returns:
            None
        """
        # List of existing datasets in the cloud.
        existing_datasets = self.get_existing_datasets()

        # Check if requested dataset exists in the cloud.
        assert (
            self.name in existing_datasets
        ), "Dataset {} does not exist in the cloud.".format(
            self.name
        ) + "Available datasets are: {}.".format(
            existing_datasets
        )

        # If the dataset does not exist locally, create the dataset folder.
        if not self._does_dataset_exist_locally():
            self._create_dataset_folder()

        # Download the dataset (metadata.json, data.pt, labels.pt)
        # by using the public URL.
        self._data_cloud.download_file(self.name + "/metadata.json")
        # load the metadata.json file to get the filetype
        with open(
            join(self.download_directory, self.name, "metadata.json")  # type: ignore
        ) as f:
            metadata = json.load(f)
        # filetype: Literal['pt', 'npy']
        if metadata["data_format"] == "pytorch_tensor":
            filetype = "pt"
        elif metadata["data_format"] == "numpy_array":
            filetype = "npy"
        else:
            raise ValueError(f"Unknown data format: {metadata['data_format']}")
        self._data_cloud.download_file(self.name + "/data." + filetype)
        self._data_cloud.download_file(self.name + "/labels." + filetype)

    def get_existing_datasets(self) -> List[str]:
        """Returns a list of datasets in the cloud.

        Returns:
            List[str]:
                List of datasets in the cloud.
        """
        if self.use_public_access:
            datasets_local = "tmp_datasets.json"
            # Download the dataset list json file using the public URL.
            wget.download(self.public_url + "datasets.json", datasets_local)  # type: ignore
            datasets = json.load(open(datasets_local))

            # Remove duplicates. This has to be fixed in the future.
            datasets = list(set(datasets))

            # Remove the temporary file.
            remove(datasets_local)

            return datasets
        else:
            existing_datasets = [
                blob_name.split("/")[0]
                for blob_name in self._data_cloud.list_blobs()
                if blob_name != "giotto-deep-big.png" and blob_name != "datasets.json"
            ]
            # Remove duplicates.
            existing_datasets = list(set(existing_datasets))

            # Remove dataset that are not public, i.e. start with "private_".
            existing_datasets = [
                dataset
                for dataset in existing_datasets
                if not dataset.startswith("private_")
            ]

            return existing_datasets

    def _update_dataset_list(self) -> None:
        """Updates the dataset list in the datasets.json file.

        Returns:
            None
        """
        self._check_public_access()

        # List of existing datasets in the cloud.
        existing_datasets = self.get_existing_datasets()

        # Save existing datasets to a json file.
        json_file = "tmp_datasets.json"
        json.dump(existing_datasets, open(json_file, "w"))

        # Upload the json file to the cloud.
        self._data_cloud.upload_file(
            json_file,
            "datasets.json",
            make_public=True,
            overwrite=True,
        )

        # Remove the temporary file.
        remove(json_file)

    @staticmethod
    def _get_filetype(path: str) -> str:
        """Returns the file extension from a given path.

        Args:
            path:
                A string path.

        Returns:
            str:
                The file extension.

        Raises:
            None.
        """
        return path.split(".")[-1]

    def _check_public_access(self) -> None:
        """Check if use_public_access is set to False."""
        assert (
            self.use_public_access is False
        ), "Only download functionality is supported for public access."

    def _upload_data(
        self,
        path: str,
    ) -> None:
        """Uploads the data file to a Cloud Storage bucket.

        Args:
        path:
            The path to the data file.

        Raises:
        ValueError:
            If the file type is not supported.

        Returns:
            None
        """
        self._check_public_access()

        filetype = DatasetCloud._get_filetype(path)

        # Check if the file type is supported
        if filetype in ["pt", "npy"]:
            self._data_cloud.upload_file(
                path,
                (self.metadata["name"] + "/data." + filetype),  # type: ignore
                make_public=self.make_public,
                overwrite=False,
            )
        else:
            raise ValueError("File type {} is not supported.".format(filetype))

    def _upload_label(
        self,
        path: str,
    ) -> None:
        """Uploads a set of labels to a remote dataset.

        Args:
            path:
                the path to the labels file.

        Raises:
            ValueError:
                If the file type is not supported.

        Returns:
            None

        """
        self._check_public_access()

        filetype = DatasetCloud._get_filetype(path)

        # Check if the file type is supported
        if filetype in ["pt", "npy"]:
            self._data_cloud.upload_file(
                path,
                (self.metadata["name"] + "/labels." + filetype),  # type: ignore
                make_public=self.make_public,
                overwrite=False,
            )
        else:
            raise ValueError("File type {} is not supported.".format(filetype))

    def _upload_metadata(self, path: Union[str, None] = None) -> None:
        """Uploads the metadata dictionary to the location specified in the
        metadata. The metadata dictionary is generated using create_metadata.

        Args:
            path (str):
                The path to the data cloud folder. If none, path will
                be set to the default path.

        Raises:
            Exception:
                If no metadata exists, an exception will be raised.

        Returns:
            None
        """
        self._check_public_access()
        self._data_cloud.upload_file(
            path,  # type: ignore
            str(self.metadata["name"]) + "/" + "metadata.json",  # type: ignore
            make_public=self.make_public,  # type: ignore
        )

    def _add_metadata(
        self,
        size_dataset: int,
        input_size: Tuple[int, ...],
        num_labels: Union[None, int] = None,
        data_type: str = "tabular",
        task_type: str = "classification",
        name: Union[None, str] = None,
        data_format: Union[None, str] = None,
        comment: Union[None, str] = None,
    ) -> None:
        """This function accepts various metadata for the dataset and stores it
        in a temporary JSON file.

        Args:
            size_dataset (int):
                The size of the dataset (in terms of the number
                of samples).
            input_size (Tuple[int, ...]):
                The size of each sample in the
                dataset.
            num_labels (Union[None, int]):
                The number of classes in the dataset.
            data_type (str):
                The type of data in the dataset.
            task_type (str):
                The task type of the dataset.
            name (Union[None, str]):
                The name of the dataset.
            data_format (Union[None, str]):
                The format of the data in the dataset.
            comment (Union[None, str]):
                A comment describing the dataset.

        Returns:
            None
        """
        self._check_public_access()
        if name is None:
            name = self.name
        if data_format is None:
            data_format = "pytorch_tensor"
        self.path_metadata = "tmp_metadata.json"  # type: ignore
        self.metadata = {
            "name": name,
            "size": size_dataset,
            "input_size": input_size,
            "num_labels": num_labels,
            "task_type": task_type,
            "data_type": data_type,
            "data_format": data_format,
            "comment": comment,
        }
        with open(self.path_metadata, "w") as f:  # type: ignore
            json.dump(self.metadata, f, sort_keys=True, indent=4)

    def _upload(
        self,
        path_data: str,
        path_label: str,
        path_metadata: Union[str, None] = None,
    ) -> None:
        """Uploads a dataset to the cloud.

        Args:
            path_data (str): Path to the data files.
            path_label (str): Path to the label file.
            path_metadata (Optional[str]): Path to the metadata file.

        Raises:
            ValueError: If the dataset already exists in the cloud.

        Returns:
            None
        """
        self._check_public_access()

        # List of existing datasets in the cloud.
        existing_datasets = self.get_existing_datasets()
        if self.name in existing_datasets:
            raise ValueError(
                "Dataset {} already exists in the cloud.".format(self.name)
                + "Available datasets are: {}.".format(existing_datasets)
            )
        if path_metadata is None:
            path_metadata = self.path_metadata
        self._upload_metadata(path_metadata)
        self._upload_data(path_data)
        self._upload_label(path_label)

        # Update dataset list.
        self._update_dataset_list()



from gdeep.data.datasets import (
    ImageClassificationFromFiles,
    DataLoaderBuilder,
    DlBuilderFromDataCloud,
)

from gdeep.data import TransformingDataset
from gdeep.data.preprocessors import ToTensorImage
import os
from os.path import join
import numpy as np
import torch
import logging
from google.auth.exceptions import DefaultCredentialsError  # type: ignore

LOGGER = logging.getLogger(__name__)


def test_images_from_file():
    """test DatasetImageClassificationFromFiles"""
    file_path = os.path.dirname(os.path.realpath(__file__))
    transform = ToTensorImage((32, 32))  # type: ignore # this is already fitted
    ds = ImageClassificationFromFiles(
        os.path.join(file_path, "img_data"),
        os.path.join(file_path, "img_data", "labels.csv"),
    )

    tds = TransformingDataset(ds, transform)
    dl, *_ = DataLoaderBuilder((tds,)).build(({"batch_size": 2},))  # type: ignore
    assert len(next(iter(dl))[0].shape) == 4


def test_dlbuilderfromdatacloud():
    dataset_name = "SmallDataset"
    download_directory = join("examples", "data", "DatasetCloud")

    dl_cloud_builder = DlBuilderFromDataCloud(dataset_name, download_directory)

    train_dataloader, val_dataloader, test_dataloader = dl_cloud_builder.build(
        ({"batch_size": 10},)  # type: ignore
    )
    x, y = next(iter(train_dataloader))
    assert x.shape == torch.Size([10, 5])
    assert y.shape == torch.Size([10])



from .simple_nn import FFNet
from .utils import SaveOutput, SaveNodeOutput, SaveLayerOutput
from .periodic_nn import PeriodicNeuralNetwork
from .extractor import ModelExtractor

__all__ = [
    "FFNet",
    "SaveOutput",
    "SaveNodeOutput",
    "SaveLayerOutput",
    "PeriodicNeuralNetwork",
    "ModelExtractor",
]



import warnings
from typing import Tuple, Any, Dict, Callable, Type, Optional, List, Union

import torch
from torch.utils.data import DataLoader
from torch.utils.tensorboard.writer import SummaryWriter
from sklearn.model_selection._split import BaseCrossValidator  # noqa
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler  # noqa
from optuna.trial._base import BaseTrial  # noqa
from sklearn.model_selection import KFold

from gdeep.trainer import Trainer
from gdeep.utility import _are_compatible  # noqa
from gdeep.trainer import accuracy, TrainerConfig
from .hpo_config import HPOConfig

from gdeep.utility.custom_types import Tensor


class Benchmark:
    """This is the generic class that allows
    the user to perform benchmarking over different
    datasets and models.

    Args:
        models_dicts :
            each dictionary has two items, ``"name":"string"``
            the name of the model and `"model":nn.Module` a
            standard torch model
        dataloaders_dicts :
            each dictionary has two items, ``"name":"string"``
            the name of the model and ``"dataloaders":list`` a
            list of standard torch.dataloaders, e.g. ``(dl_tr, dl_ts)``
        loss_fn:
            loss function
        writer:
            tensorboard writer
        training_metric:
            the function that computes the metric: it shall
            have two arguments, one for the prediction
            and the other for the ground truth
        k_fold_class:
            the class instance to implement the KFold, can be
            any of the Splitter classes of sklearn. More
            info at https://scikit-learn.org/stable/modules/classes.html#module-sklearn.model_selection

    Examples::

        from torch import nn
        from torch.optim import SGD
        from torch.utils.data.sampler import SubsetRandomSampler
        from gdeep.data.datasets import DatasetBuilder, DataLoaderBuilder
        from gdeep.data.preprocessors import ToTensorImage
        from gdeep.search import Benchmark
        from gdeep.models import FFNet
        # initialise dataloader list of dictionaries
        dataloaders_dicts = []
        bd = DatasetBuilder(name="CIFAR10")
        # build first dataloader
        ds_tr, *_ = bd.build()
        transformation = ToTensorImage((32, 32))
        transformed_ds_tr = transformation.attach_transform_to_dataset(ds_tr)
        # cut down some data
        test_indices = [64 * 5 + x for x in range(32 * 3)]
        train_indices = [x for x in range(32 * 2)]
        dl = DataLoaderBuilder((transformed_ds_tr, transformed_ds_tr))
        dl_tr, dl_val, _ = dl.build(
            (
                {"batch_size": 32, "sampler": SubsetRandomSampler(train_indices)},
                {"batch_size": 32, "sampler": SubsetRandomSampler(test_indices)},
            )
        )
        # prepare the dataloader dictionary
        temp_dict = {}
        temp_dict["name"] = "CIFAR10_1000"
        temp_dict["dataloaders"] = (dl_tr, dl_val, _)
        # store the dictionary to the list
        dataloaders_dicts.append(temp_dict)
        # repeat with another dataset
        db = DatasetBuilder(name="DoubleTori")
        ds_tr, ds_val, _ = db.build()
        dl_tr, dl_ts, _ = DataLoaderBuilder((ds_tr, ds_val)).build(
            ({"batch_size": 48}, {"batch_size": 32})
        )
        temp_dict = {}
        temp_dict["name"] = "double_tori"
        temp_dict["dataloaders"] = (dl_tr, dl_ts)
        dataloaders_dicts.append(temp_dict)
        # prepare the list of model dictionaries
        models_dicts = []
        # define your model as a torch.nn.Module
        model = model2()
        temp_dict = {}
        temp_dict["name"] = "resnet18"
        temp_dict["model"] = model
        models_dicts.append(temp_dict)
        # avoid having exposed paramters that wll not be searched on
        class model_no_param(nn.Module):
            def __init__(self):
                super(model_no_param, self).__init__()
                self.mod = FFNet([3, 5, 5, 2])

            def forward(self, x):
                return self.mod(x)
        # initialise the dictionary of the other model
        model5 = model_no_param()
        temp_dict = {}
        temp_dict["name"] = "ffnn"
        temp_dict["model"] = model5
        # append to the model list of dictionaries
        models_dicts.append(temp_dict)
        # standard pytorch loss
        loss_fn = nn.CrossEntropyLoss()

        # initialise benchmark
        bench = Benchmark(
            models_dicts, dataloaders_dicts, loss_fn, writer, k_fold_class=KFold(3)
        )
        # start the benchmarking
        bench.start(SGD, 1, True, {"lr": 0.01}, {"batch_size": 23})

    """

    def __init__(
        self,
        models_dicts: List[Dict[str, Union[torch.nn.Module, str]]],
        dataloaders_dicts: List[
            Dict[str, Union[List[DataLoader[Tuple[Tensor, Tensor]]], str]]
        ],
        loss_fn: Callable[[Tensor, Tensor], Tensor],
        writer: SummaryWriter,
        training_metric: Optional[Callable[[Tensor, Tensor], float]] = None,
        k_fold_class: Optional[BaseCrossValidator] = None,
    ) -> None:
        self.models_dicts = models_dicts
        self.dataloaders_dicts = dataloaders_dicts
        self.loss_fn = loss_fn
        self.writer = writer
        if not self.writer:
            warnings.warn("No writer detected")

        if not k_fold_class:
            self.k_fold_class = KFold(5, shuffle=True)
        else:
            self.k_fold_class = k_fold_class
        if training_metric:
            self.training_metric = training_metric
        else:
            self.training_metric = accuracy
        if not isinstance(self.models_dicts, list):
            raise TypeError("The provided models must be a Python list of dictionaries")

        if not isinstance(self.dataloaders_dicts, list):
            raise TypeError(
                "The provided datasets must be a Python list of dictionaries"
            )

    def start(
        self,
        optimizer: Type[Optimizer],
        n_epochs: int = 10,
        cross_validation: bool = False,
        optimizers_param: Optional[Dict[str, Any]] = None,
        dataloaders_param: Optional[Dict[str, Any]] = None,
        lr_scheduler: Optional[Type[_LRScheduler]] = None,
        scheduler_params: Optional[Dict[str, Any]] = None,
        profiling: bool = False,
        parallel_tpu: bool = False,
        keep_training: bool = False,
        store_grad_layer_hist: bool = False,
        n_accumulated_grads: int = 0,
        writer_tag: str = "",
    ) -> None:
        """Method to be called when starting the benchmarking

        Args:
            optimizer:
                a torch optimizers class (not the instance)
            n_epochs:
                number of training epochs
            cross_validation:
                whether or not to use cross-validation
            optimizers_param:
                dictionary of the optimizers
                parameters, e.g. ``{"lr": 0.001}``
            dataloaders_param:
                dictionary of the dataloaders
                parameters, e.g. ``{"batch_size": 32}``
            lr_scheduler:
                a learning rate scheduler class (not instance)
            scheduler_params:
                learning rate scheduler parameters
            profiling:
                whether or not you want to activate the
                profiler
            parallel_tpu:
                boolean value to run the computations
                on multiple TPUs
            keep_training:
                This flag allows to restart a training from
                the existing optimizer as well as the
                existing model
            store_grad_layer_hist:
                This flag allows to store the gradients
                and the layer values in tensorboard for
                each epoch
            n_accumulated_grads:
                number of accumulated gradients. It is
                considered only if a positive integer
            writer_tag:
                string to be added to the tensorboard items title
        """

        print("Benchmarking Started")
        config = TrainerConfig(
            optimizer,
            n_epochs,
            cross_validation,
            optimizers_param,
            dataloaders_param,
            lr_scheduler,
            scheduler_params,
            None,
            profiling,
            parallel_tpu,
            keep_training,
            store_grad_layer_hist,
            n_accumulated_grads,
            writer_tag,
        )

        _benchmarking_param(
            self._inner_function, (self.models_dicts, self.dataloaders_dicts), config  # type: ignore
        )

    def _inner_function(
        self,
        model: Dict[str, Union[torch.nn.Module, str]],
        dataloaders: Dict[str, Union[List[DataLoader[Tuple[Tensor, Tensor]]], str]],
        config: TrainerConfig,
    ) -> None:
        """private method to run the inner
        function of the benchmark loops

        Args:
            model:
                dictionary defining the model name
                and actual nn.Module
            dataloaders:
                dictionary defining the dataset name
                and the actual list of dataloaders, e.g.
                ``[dl_tr, dl_val, dl_ts]``
            config:
                the configuration class ``TrainerConfig``.
        """
        pipe = Trainer(
            model["model"],  # type: ignore
            dataloaders["dataloaders"],  # type: ignore
            self.loss_fn,
            self.writer,
            self.training_metric,  # type: ignore
            self.k_fold_class,
        )
        config.writer_tag += "Dataset:" + dataloaders["name"] + "|Model:" + model["name"]  # type: ignore
        pipe.train(**config.to_dict())


def _benchmarking_param(
    fun: Callable[[Any, Any, Any], Any],
    arguments: Tuple[
        List[Dict[str, Union[torch.nn.Module, str]]],
        List[Dict[str, Union[List[DataLoader[Tuple[Tensor, Tensor]]], str]]],
    ],
    config: Union[HPOConfig, TrainerConfig],
) -> None:
    """Function to be used as pseudo-decorator for
    benchmarking loops

    Args:
        fun (Callable):
            the function to decorate
        arguments (list):
            list of arguments to pass to the inner
            function of the wrapper. Expected to receive
            ``models_dicts, dataloaders_dict = arguments``
        config:
            the configuration class ``TrainerConfig``.

    """

    models_dicts, dataloaders_dicts = arguments
    for dataloaders in dataloaders_dicts:
        for model in models_dicts:
            if _are_compatible(model, dataloaders):
                print("*" * 40)
                print(
                    f"Performing Gridsearch on Dataset: {dataloaders['name']}"
                    f", Model: {model['name']}"
                )
                fun(model, dataloaders, config)



# Joint with Matthias Kemper

import torch
import torch.optim
from typing import Callable
from abc import ABC, abstractmethod
from torchdiffeq import odeint

if torch.cuda.is_available():
    DEVICE = torch.device("cuda")
else:
    DEVICE = torch.device("cpu")


class DecisionBoundaryCalculator(ABC):
    """
    Abstract class for calculating the decision boundary of
    a neural network
    """

    @abstractmethod
    def step(self, number_of_steps=1):
        """Performs a single step towards the decision boundary"""
        pass

    @abstractmethod
    def get_decision_boundary(self) -> torch.Tensor:
        """Return current state and does not make a step

        Returns:
            torch.tensor:
                current estimate of the decision boundary
        """
        pass

    @abstractmethod
    def get_filtered_decision_boundary(self, delta=0.01) -> torch.Tensor:
        pass

    def _convert_to_distance_to_boundary(self, model, output_shape):
        """Convert the binary or multiclass classifier to a scalar valued model with
        distance to the decision boundary as output.
        """
        if len(output_shape) == 1:
            new_model = lambda x: torch.abs(model(x) - 0.5)
        elif output_shape[-1] == 1:
            new_model = lambda x: torch.abs(model(x).reshape((-1)) - 0.5)
        elif output_shape[-1] == 2:
            new_model = lambda x: torch.abs(model(x)[:, 0] - 0.5)
        else:

            def new_model(x):
                y = torch.topk(model(x), 2).values
                return y[:, 0] - y[:, 1]

        return new_model


class GradientFlowDecisionBoundaryCalculator(DecisionBoundaryCalculator):
    """
    Computes Decision Boundary using the gradient flow method

    Args:
        model (Callable[[torch.Tensor], torch.Tensor]):
            Function that maps
            a ``torch.Tensor`` of shape (N, D_in) to a tensor either of
            shape (N) and with values in [0,1] or of shape (N, D_out) with
            values in [0, 1] such that the last axis sums to 1.
        initial_points (torch.Tensor):
            ``torch.Tensor`` of shape (N, D_in)
        optimizer (Callable[[torch.Tensor], torch.optim.Optimizer]):
            Function returning an optimizer for the params given as an
            argument.
    """

    def __init__(
        self,
        model: Callable[[torch.Tensor], torch.Tensor],
        initial_points: torch.Tensor,
        optimizer: Callable[[list], torch.optim.Optimizer],
    ):

        self.sample_points = initial_points.to(DEVICE)
        self.sample_points.requires_grad = True
        output = model(self.sample_points).to(DEVICE)
        output_shape = output.size()

        if not len(output_shape) in [1, 2]:
            raise RuntimeError("Output of model has wrong size!")
        # Convert `model` to `self.model` with squared generalized distance
        # to the decision boundary as output.
        new_model = self._convert_to_distance_to_boundary(model, output_shape)
        self.model = lambda x: new_model(x) ** 2
        # Check if self.model has the right output shape
        assert (
            len(self.model(self.sample_points).size()) == 1
        ), f"Output shape is {self.model(self.sample_points).size()}"

        self.optimizer = optimizer([self.sample_points])

    def step(self, number_of_steps=1):
        """Performs the indicated number of steps towards the decision boundary"""
        print("Executing the decision boundary computations:")
        for j in range(number_of_steps):
            print("Step: " + str(j) + "/" + str(number_of_steps), end="\r")
            self.optimizer.zero_grad()
            loss = torch.sum(self.model(self.sample_points)).to(DEVICE)
            loss.backward()
            # if DEVICE.type == "xla":
            #    xm.optimizer_step(self.optimizer, barrier=True)  # Note: Cloud TPU-specific code!
            # else:
            self.optimizer.step()

    def get_decision_boundary(self) -> torch.Tensor:
        return self.sample_points

    def get_filtered_decision_boundary(self, delta=0.01) -> torch.Tensor:
        q_dist = self.model(self.sample_points)
        return self.sample_points[q_dist <= delta**2]


class QuasihyperbolicDecisionBoundaryCalculator(DecisionBoundaryCalculator):
    """
    Computes Decision Boundary by emanating quasihyperbolic geodesics
    """

    def __init__(
        self,
        model: Callable[[torch.Tensor], torch.Tensor],
        initial_points: torch.Tensor,
        initial_vectors: torch.Tensor,
        integrator=None,
    ):
        """
        Args:
            model (Callable[[torch.Tensor], torch.Tensor]):
                Function that maps
                a `torch.Tensor` of shape (N, D_in) to a tensor either of
                shape (N) and with values in [0,1] or of shape (N, D_out) with
                values in [0, 1] such that the last axis sums to 1.

            initial_points (torch.Tensor):
                `torch.Tensor` of shape (N, D_in)
                containing the starting points.

            initial_vectors(torch.Tensor):
                `torch.Tensor` of shape (N, D_in)
                containing the starting tangent vectors (directions).
                Prefarably normalized.

            integrator:
                unused
        """
        self.points = initial_points

        output = model(self.points)
        output_shape = output.size()

        if not len(output_shape) in [1, 2]:
            raise RuntimeError("Output of model has wrong size!")
        # Convert `model` to `self.model` with generalized distance
        # to the decision boundary as output.
        self.model = self._convert_to_distance_to_boundary(model, output_shape)

        # Check if `self.model` has the right output shape
        distance_function = self.model(self.points)
        assert (
            len(distance_function.size()) == 1
        ), f"Output shape is {distance_function.size()}"
        # Normalize tangent vectors in quasihyperbolic metric
        self.vectors = initial_vectors * distance_function.unsqueeze(-1)

        self.integrator = integrator

    def step(self, number_of_steps=1):
        """Performs the indicated number of steps towards
        the decision boundary
        """

        # Calculate logarithmic gradient of generalized
        # distance function at points
        def gradient(y):
            # self.points.grad.zero_()
            # loss = torch.sum(torch.log(self.model(self.points)))
            # loss.backward()
            y.requires_grad = True
            loss = torch.sum(torch.log(self.model(y)))
            loss.backward()
            return y.grad.detach()

        # quasi-hyperbolic geodesic equation
        def odes(t, x):
            y = x[0]
            dy = x[1]

            gradient_log_delta = gradient(y)

            # quasi-hyperbolic geodesic equation see markdown comment
            ddy = 2 * torch.einsum(
                "bi,bi,bj->bj", gradient_log_delta, dy, dy
            ) - torch.einsum("bi,bj,bj->bi", gradient_log_delta, dy, dy)
            return torch.stack((dy, ddy))

        ode_initial_conditions = torch.stack((self.points, self.vectors))

        step_size = 1e-1
        # t = torch.arange(0, 3e0, 2e-3)

        ode_solver_output = odeint(
            odes,
            ode_initial_conditions,
            torch.tensor(
                [0, number_of_steps * step_size],
                dtype=self.points.dtype,
                device=self.points.device,
            ),
            method="rk4",
            options=dict(step_size=step_size),
        )
        self.points = ode_solver_output[1, 0]  # type: ignore
        self.vectors = ode_solver_output[1, 1]  # type: ignore

    def get_decision_boundary(self) -> torch.Tensor:
        return self.points

    def get_filtered_decision_boundary(self, delta=0.01) -> torch.Tensor:
        distance = self.model(self.points)
        return self.points[distance <= delta]



from inspect import signature
from typing import Any, Callable, List, Tuple, Union

import torch
from numpy import ndarray

from gdeep.utility.custom_types import Tensor


def torch_transform(
    transform: Union[Callable[[Tensor], Tensor], Callable[[ndarray], ndarray]]
) -> Callable[[Tensor], Tensor]:
    """Transforms a numpy array transform to a torch transform.

    Args:
        transform: Either a callable that takes a numpy array and returns a
            numpy array or a callable that takes a tensor and returns a tensor.

    Returns:
        The torch transform.
    """
    if (
        get_parameter_types(transform)[0] is Tensor
        and get_return_type(transform) is Tensor
    ):
        return transform  # type: ignore
    elif (
        get_parameter_types(transform)[0] is ndarray
        and get_return_type(transform) is ndarray
    ):
        return lambda x: torch.tensor(transform(x.numpy()))
    else:
        raise ValueError(
            "Transform must be a function that takes a tensor or"
            "an array and returns a tensor or an array."
            "Please provide type annotations."
        )


def get_parameter_types(func: Callable) -> List[Any]:
    """Returns a list of the types of the parameters of a function.

    Args:
        func: The function to get the types of the parameters of.
    Returns:
        A list of the types of the parameters of the function.
    """
    return [t.annotation for t in signature(func).parameters.values()]


def get_return_type(func: Callable) -> Any:
    """Returns the type of the return value of a function.

    Args:
        func: The function to get the type of the return value of.

    Returns:
        The type of the return value of the function.
    """
    return signature(func).return_annotation



from dataclasses import dataclass
import json
import os
import shutil
import warnings
from abc import ABC, abstractmethod
from os.path import join
from collections.abc import Iterable
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    List,
)

import torch
from torch.utils.data import DataLoader, Dataset
from torchvision.transforms import Resize, ToTensor
from tqdm import tqdm
from torch.utils.data import Sampler

from .build_datasets import get_dataset
from .dataset_cloud import DatasetCloud
from ..transforming_dataset import TransformingDataset


from gdeep.utility.custom_types import Tensor

T = TypeVar("T")


class AbstractDataLoaderBuilder(ABC):
    """The abstract class to interface the
    Giotto dataloaders"""

    @abstractmethod
    def build(self, tuple_of_kwargs: List[Dict[str, Any]]):
        pass


@dataclass
class DataLoaderParams:
    batch_size: Optional[int] = 1
    shuffle: bool = False
    num_workers: int = 0
    collate_fn: Optional[Callable[[Any], Any]] = None
    batch_sampler: Optional[Sampler[Sequence]] = None
    pin_memory: bool = False
    drop_last: bool = False
    timeout: float = 0.0
    persistent_workers: bool = False

    def copy(self):
        return DataLoaderParams(**self.to_dict())

    def update_batch_size(self, batch_size: int):
        self.batch_size = batch_size
        return self

    def update_shuffle(self, shuffle: bool):
        self.shuffle = shuffle
        return self

    def to_dict(self):
        return {k: v for k, v in self.__dict__.items() if not k.startswith("_")}


@dataclass
class DataLoaderParamsTuples:
    train: DataLoaderParams
    test: Optional[DataLoaderParams] = None
    validation: Optional[DataLoaderParams] = None

    @staticmethod
    def default(
        collate_fn: Callable[[Any], Any],
        batch_size: int = 32,
        num_workers: int = 0,
        pin_memory: bool = False,
        drop_last: bool = False,
        timeout: float = 0.0,
        persistent_workers: bool = False,
        with_validation: bool = False,
    ) -> "DataLoaderParamsTuples":
        dlp = DataLoaderParamsTuples(
            train=DataLoaderParams(
                batch_size=batch_size,
                shuffle=True,
                num_workers=num_workers,
                collate_fn=collate_fn,
                pin_memory=pin_memory,
                drop_last=drop_last,
                timeout=timeout,
                persistent_workers=persistent_workers,
            ),
            test=DataLoaderParams(
                batch_size=batch_size,
                shuffle=False,
                num_workers=num_workers,
                collate_fn=collate_fn,
                pin_memory=pin_memory,
                drop_last=drop_last,
                timeout=timeout,
                persistent_workers=persistent_workers,
            ),
        )
        if with_validation:
            dlp.validation = DataLoaderParams(
                batch_size=batch_size,
                shuffle=False,
                num_workers=num_workers,
                collate_fn=collate_fn,
                pin_memory=pin_memory,
                drop_last=drop_last,
                timeout=timeout,
                persistent_workers=persistent_workers,
            )
        return dlp

    def to_tuple_of_dicts(self) -> Tuple[Dict, ...]:
        if self.validation is not None and self.test is not None:
            return (
                self.train.to_dict(),
                self.validation.to_dict(),
                self.test.to_dict(),
            )
        elif self.test is not None:
            return (self.train.to_dict(), self.test.to_dict())
        else:
            return (self.train.to_dict(),)

    @staticmethod
    def from_list_of_dicts(
        list_of_kwargs: List[Dict[str, Any]]
    ) -> "DataLoaderParamsTuples":
        """This method accepts the arguments of the torch
        Dataloader and applies them when creating the
        tuple

        Args:
            list_of_kwargs:
                List of dictionaries, each one being the
                kwargs for the corresponding DataLoader
        """
        if len(list_of_kwargs) == 3:
            return DataLoaderParamsTuples(
                DataLoaderParams(**list_of_kwargs[0]),
                DataLoaderParams(**list_of_kwargs[1]),
                DataLoaderParams(**list_of_kwargs[2]),
            )
        elif len(list_of_kwargs) == 2:
            return DataLoaderParamsTuples(
                DataLoaderParams(**list_of_kwargs[0]),
                DataLoaderParams(**list_of_kwargs[1]),
                None,
            )
        elif len(list_of_kwargs) == 1:
            return DataLoaderParamsTuples(
                DataLoaderParams(**list_of_kwargs[0]), None, None
            )
        else:
            raise ValueError("The list of dictionaries must have 2 or 3 elements")


class DataLoaderBuilder(AbstractDataLoaderBuilder):
    """This class builds, out of a tuple of datasets, the
    corresponding dataloaders. Note that this class would
    use the same parameters for all the datasets. You can
    use different parameters for each dataset by passing
    a list of dictionaries to the build method

    Args:
        tuple_of_datasets :
            Tuple consisting of the training, validation and test
            datasets. Also one or two elements are acceptable:
            they will be considered as training first and
            validation afterwards.

    Example:
        >>> import torch
        >>> from gdeep.data.dataloaders import DataLoaderBuilder
        >>> x, y = torch.rand(10, 3, 32, 32), torch.randint(0, 1, (10,))
        >>> x_train, y_train = x[:8], y[:8]
        >>> x_val, y_val = x[8:], y[8:]
        >>> train_dataset = gdeep.data.datasets.FromArray(x_train, y_train)
        >>> val_dataset = gdeep.data.datasets.FromArray(x_val, y_val)
        >>> dataloader_builder = DataLoaderBuilder(train_dataset, val_dataset)
        >>> train_loader, val_loader = dataloader_builder.build()
    """

    def __init__(self, tuple_of_datasets: List[Dataset[Any]]) -> None:
        self.tuple_of_datasets = tuple_of_datasets
        assert len(tuple_of_datasets) <= 3, "Too many Dataset inserted: maximum 3."

    def build(
        self,
        tuple_of_kwargs: Union[
            List[Dict[str, Any]], DataLoaderParamsTuples, None
        ] = None,
    ) -> List[DataLoader[Any]]:
        """This method accepts the arguments of the torch
        Dataloader and applies them when creating the
        tuple. If the tuple of kwargs is a list of dictionaries,
        then the first dictionary will be applied to the
        training dataset, the second to the validation dataset
        and the third to the test dataset. If the tuple of
        kwargs is a DataLoaderParamsTuples, then the parameters
        will be applied to the corresponding dataset.

        Args:
            tuple_of_kwargs:
                Tuple consisting of the training, validation and test
                dataloaders. Also one or two elements are acceptable:
                they will be considered as training first and
                validation afterwards.
        """
        out: List = []
        if tuple_of_kwargs is None:
            for i, dataset in enumerate(self.tuple_of_datasets):
                out.append(DataLoader(dataset))
            out += [None] * (3 - len(out))
            return out

        if isinstance(tuple_of_kwargs, DataLoaderParamsTuples):
            tuple_of_kwargs = tuple_of_kwargs.to_tuple_of_dicts()  # type: ignore

        assert isinstance(tuple_of_kwargs, (list, tuple)), (
            "The kwargs must be a list or a tuple at" "this point"
        )
        assert len(tuple_of_kwargs) == len(
            self.tuple_of_datasets
        ), "Cannot match the dataloaders and the parameters. "
        for dataset, kwargs in zip(self.tuple_of_datasets, tuple_of_kwargs):
            out.append(DataLoader(dataset, **kwargs))
        out += [None] * (3 - len(out))
        return out



import math
from typing import Tuple, Union, List

import numpy as np
from sklearn.datasets import make_blobs
import torch
from torch.utils.data import Dataset


from gdeep.utility.custom_types import Tensor

from gdeep.utility.custom_types import Array


class Rotation:
    """Class for rotations"""

    def __init__(self, axis_0: int, axis_1: int, angle: float) -> None:
        self._axis_0 = axis_0
        self._axis_1 = axis_1
        self._angle = angle

    def return_axis(self, idx: int):
        return eval("self._axis_" + str(idx))

    def return_angle(self) -> float:
        return self._angle

    def rotation_matrix(self) -> Array:
        rotation_matrix = np.identity(3)
        rotation_matrix[self._axis_0, self._axis_0] = math.cos(self._angle)
        rotation_matrix[self._axis_1, self._axis_1] = math.cos(self._angle)
        rotation_matrix[self._axis_1, self._axis_0] = math.sin(self._angle)
        rotation_matrix[self._axis_0, self._axis_1] = -math.sin(self._angle)
        return rotation_matrix


class ToriDataset(Dataset[Tuple[Tensor, Tensor]]):
    """This class is used to generate data loaders for the
    family of tori-datasets

    Args:
        name:
            name of the torus dataset to generate
    """

    def __init__(self, name: str, **kwargs) -> None:
        self.name = name
        if self.name == "DoubleTori":
            self.out_tuple = self._make_two_tori_dataset(**kwargs)
        elif self.name == "Blobs":
            self.out_tuple = self._make_blobs(**kwargs)
        else:
            self.out_tuple = self._make_entangled_tori_dataset(**kwargs)

    def __getitem__(self, idx: int) -> Tuple[Tensor, Tensor]:
        """This method is the main method of the class.
        It extracts the i-the element
        """

        return self.out_tuple[0][idx], torch.tensor(self.out_tuple[1][idx]).to(
            torch.long
        )

    def __len__(self) -> int:
        return len(self.out_tuple[0])

    @staticmethod
    def _make_torus_point_cloud(
        label: int,
        n_points: int,
        noise: float,
        rotation: Rotation,
        base_point: Array,
        radius: float = 1.0,
    ) -> Tuple[Array, Array]:
        """Generate point cloud of a single torus

        Args:
            label (int):
                label of the data points
            n_points (int):
                number of sample points for each direction
            noise (float):
                noise
            rotation (Rotation):
                rotation class
            base_point (np.array):
                center of the torus
            radius (float):
                radius of the torus

        Returns:
            (np.array, np.array):
                data_points, labels
        """
        torus_point_clouds = np.asarray(
            [
                [
                    (2 + radius * np.cos(s)) * np.cos(t)  # type: ignore
                    + noise * (np.random.rand() - 0.5),
                    (2 + radius * np.cos(s)) * np.sin(t)  # type: ignore
                    + noise * (np.random.rand() - 0.5),
                    radius * np.sin(s) + noise * (np.random.rand(1)[0] - 0.5),  # type: ignore
                ]
                for t in range(n_points)
                for s in range(n_points)
            ]
        )  # type: ignore

        torus_point_clouds = np.einsum(
            "ij,kj->ki", rotation.rotation_matrix(), torus_point_clouds
        )

        torus_point_clouds += base_point

        # label tori with 2
        torus_labels = label * np.ones(n_points**2)

        return torus_point_clouds, torus_labels

    @staticmethod
    def _make_torus_point_cloud2(
        label: int,
        n_points: int,
        noise: float,
        rotation: Rotation,
        base_point: Union[Array, List],
        radius1: float = 1.0,
        radius2: float = 1.0,
    ) -> Tuple[Array, Array]:
        """Generate point cloud of a single torus using
        2 radii for its definition

        Args:
            label :
                label of the data points
            n_points :
                number of sample points for each direction
            noise :
                noise
            rotation:
                Rotation
            base_point :
                center of the torus
            radius1 :
                radius of torus 1
            radius2 :
                radius of torus 2

        Returns:
            (np.ndarray, np.ndarray):
                data_points, labels
        """
        torus_point_clouds = np.asarray(
            [
                [
                    (radius1 + radius2 * np.cos(s)) * np.cos(t)  # type: ignore
                    + noise * (np.random.rand() - 0.5),
                    (radius1 + radius2 * np.cos(s)) * np.sin(t)  # type: ignore
                    + noise * (np.random.rand() - 0.5),
                    radius2 * np.sin(s) + noise * (np.random.rand() - 0.5),  # type: ignore
                ]
                for t in range(n_points)
                for s in range(n_points)
            ]
        )  # type: ignore

        torus_point_clouds = np.einsum(
            "ij,kj->ki", rotation.rotation_matrix(), torus_point_clouds
        )

        torus_point_clouds += base_point

        # label tori with 2
        torus_labels = label * np.ones(n_points**2)

        return torus_point_clouds, torus_labels

    def _make_two_tori_dataset(
        self, entangled: bool = True, n_pts: int = 10
    ) -> Tuple[Tensor, Array]:
        """Generates pandas Dataframe of two tori in 3D. The labels correspond to
        the different Tori.

        Args:
            entangled :
                Either entangled or unentangled tori. Defaults to True.

        Returns:
            (tuple):
                the tuple for data and labels
        """
        torus_point_cloud = {}
        torus_labels = {}
        if entangled:
            torus_point_cloud[0], torus_labels[0] = self._make_torus_point_cloud(
                0,
                n_pts,
                0.0,
                Rotation(1, 2, math.pi / 2),
                np.array([[0, 0, 0]]),
                radius=0.3,
            )
            torus_point_cloud[1], torus_labels[1] = self._make_torus_point_cloud(
                1, n_pts, 0.0, Rotation(1, 2, 0), np.array([[2, 0, 0]]), radius=0.3
            )
        else:
            torus_point_cloud[0], torus_labels[0] = self._make_torus_point_cloud(
                0,
                n_pts,
                0.0,
                Rotation(1, 2, math.pi / 2),
                np.array([[0, 0, 0]]),
                radius=0.3,
            )
            torus_point_cloud[1], torus_labels[1] = self._make_torus_point_cloud(
                1, n_pts, 0.0, Rotation(1, 2, 0), np.array([[6, 0, 0]]), radius=0.3
            )

        tori_point_cloud = np.concatenate(
            (torus_point_cloud[0], torus_point_cloud[1]), axis=0
        )
        tori_labels = np.concatenate((torus_labels[0], torus_labels[1]), axis=0)
        return torch.from_numpy(tori_point_cloud).to(torch.float32), tori_labels  # type: ignore

    def _make_entangled_tori_dataset(
        self, m: int = 2, n_pts: int = 10
    ) -> Tuple[Tensor, Array]:
        """Generates pandas Dataframe of m x m x m tori in 3D.
        The labels correspond to the different Tori.

        Args:
            m :
                Number of entangled tori per axis
            n_pts :
                Number of points per torus

        Returns:
            (tuple):
                the tuple for data and labels
        """
        data1, labels1 = self._make_torus_point_cloud2(
            0, n_pts, 0.1, Rotation(0, 1, 0), [0, 0, 0], 2.0, 0.5
        )
        data2, labels2 = self._make_torus_point_cloud2(
            1, n_pts, 0.1, Rotation(1, 2, 180), [2, 0, 0], 2.0, 0.5
        )

        translations = [
            [i * 10, j * 10, k * 10]
            for i in range(m)
            for j in range(m)
            for k in range(m)
        ]

        tori_entangled = np.append(data1, data2, axis=0)
        labels = np.append(labels1, labels2)
        lab = labels
        data = tori_entangled

        for t in translations:
            if t != [0, 0, 0]:
                data = np.append(data, np.add(tori_entangled, t), axis=0)  # type: ignore
                lab = np.append(lab, labels)
        return torch.from_numpy(data).to(torch.float32), lab  # type: ignore

    @staticmethod
    def _make_blobs(m: int = 3, n_pts: int = 200) -> Tuple[Tensor, Array]:
        """Generates blobs in 3D.
        The labels correspond to the different blob.

        Args:
            m (int, defalut=2):
                Number of entangled tori per axis
            n_pts (int):
                Number of points per torus

        Returns:
            (tuple):
                the tuple for data and labels
        """
        data, lab = make_blobs(  # type: ignore
            n_samples=n_pts, centers=m, n_features=3, random_state=42
        )
        return torch.from_numpy(data).to(torch.float32), lab



import os
from typing import Tuple

import torch

from gdeep.data.datasets import ImageClassificationFromFiles
from gdeep.data.datasets.persistence_diagrams_from_files import (
    PersistenceDiagramFromFiles,
)
from gdeep.data.datasets.persistence_diagrams_from_graphs_builder import (
    PersistenceDiagramFromGraphBuilder,
)
from gdeep.data.persistence_diagrams.one_hot_persistence_diagram import (
    OneHotEncodedPersistenceDiagram,
)
from gdeep.data.preprocessors import Normalization, ToTensorImage
from gdeep.data.preprocessors.filter_persistence_diagram_by_homology_dimension import (
    FilterPersistenceDiagramByHomologyDimension,
)
from gdeep.data.preprocessors.filter_persistence_diagram_by_lifetime import (
    FilterPersistenceDiagramByLifetime,
)
from gdeep.data.preprocessors.normalization_persistence_diagram import (
    NormalizationPersistenceDiagram,
)
from gdeep.utility.constants import DEFAULT_GRAPH_DIR
from gdeep.data.preprocessors.normalization import _compute_mean_of_dataset
from sklearn import preprocessing
from sklearn.model_selection import train_test_split
from torch.utils.data import Subset

from ...preprocessing_pipeline import PreprocessingPipeline
from ...transforming_dataset import TransformingDataset


def test_preeprocessing_pipeline():
    """test preprocessing pipeline"""
    normalize = Normalization()
    to_tensor = ToTensorImage([32, 32])
    file_path = os.path.dirname(os.path.realpath(__file__))

    image_dataset = ImageClassificationFromFiles(
        os.path.join(file_path, "img_data"),
        os.path.join(file_path, "img_data", "labels.csv"),
    )
    # define preprocessing pipeline
    preprocessing_pipeline = PreprocessingPipeline((to_tensor, normalize))

    # fit preprocessing pipeline to dataset
    preprocessing_pipeline.fit_to_dataset(image_dataset)

    # apply preprocessing pipeline to dataset
    preprocessed_dataset = preprocessing_pipeline.attach_transform_to_dataset(
        image_dataset
    )

    assert len(preprocessed_dataset[0][0].shape) == 3


def test_preprocessing_pipeline_for_persistence_diagrams():
    # Parameters
    name_graph_dataset: str = "MUTAG"
    diffusion_parameter: float = 0.1
    num_homology_types: int = 4

    # Create the persistence diagram dataset
    pd_creator = PersistenceDiagramFromGraphBuilder(
        name_graph_dataset, diffusion_parameter
    )
    pd_creator.create()

    pd_mutag_ds = PersistenceDiagramFromFiles(
        os.path.join(
            DEFAULT_GRAPH_DIR, f"MUTAG_{diffusion_parameter}_extended_persistence"
        )
    )
    # Create the train/validation/test split

    train_indices, test_indices = train_test_split(
        range(len(pd_mutag_ds)),
        test_size=0.2,
        random_state=42,
    )

    train_indices, validation_indices = train_test_split(
        train_indices,
        test_size=0.2,
        random_state=42,
    )

    # Create the data loaders
    train_dataset = Subset(pd_mutag_ds, train_indices)
    validation_dataset = Subset(pd_mutag_ds, validation_indices)
    test_dataset = Subset(pd_mutag_ds, test_indices)

    # Preprocess the data
    preprocessing_pipeline = PreprocessingPipeline[
        Tuple[OneHotEncodedPersistenceDiagram, int]
    ](
        (
            FilterPersistenceDiagramByHomologyDimension[int]([0, 1]),
            FilterPersistenceDiagramByLifetime[int](
                min_lifetime=-0.1, max_lifetime=1.0
            ),
            NormalizationPersistenceDiagram[int](num_homology_dimensions=4),
        )
    )

    preprocessing_pipeline.fit_to_dataset(train_dataset)

    train_dataset = preprocessing_pipeline.attach_transform_to_dataset(train_dataset)
    validation_dataset = preprocessing_pipeline.attach_transform_to_dataset(
        validation_dataset
    )
    test_dataset = preprocessing_pipeline.attach_transform_to_dataset(test_dataset)

    computed_mean = _compute_mean_of_dataset(
        TransformingDataset(
            train_dataset, lambda x: (x[0].get_raw_data().mean(dim=0), x[1])
        )
    )[:2]
    # -> (0.0, 0.0, sth, sth)
    # print(computed_mean.float())
    assert torch.allclose(computed_mean.float(), torch.tensor([0.0, 0.0]), atol=1e-2)
    computed_stddev = _compute_mean_of_dataset(
        TransformingDataset(
            train_dataset, lambda x: ((x[0].get_raw_data() ** 2).mean(dim=0), x[1])
        )
    )[:2]
    # -> (1.0, 1.0, sth, sth)
    # print(computed_stddev.float())
    assert torch.allclose(computed_stddev.float(), torch.tensor([1.0, 1.0]), atol=1e-2)



from .attention_base import AttentionBase
from .scaled_dot_product_attention import ScaledDotProductAttention
from .induced_attention import InducedAttention


__all__ = [
    "AttentionBase",
    "ScaledDotProductAttention",
    "InducedAttention",
]



import warnings
from typing import Any, List, Tuple, TypeVar

import torch
from gdeep.data.persistence_diagrams.one_hot_persistence_diagram import (
    OneHotEncodedPersistenceDiagram,
)
from torch.utils.data import Dataset

from ..abstract_preprocessing import AbstractPreprocessing
from ..transforming_dataset import TransformingDataset

from gdeep.utility.custom_types import Tensor

PD = OneHotEncodedPersistenceDiagram

T = TypeVar("T")


class FilterPersistenceDiagramByHomologyDimension(
    AbstractPreprocessing[Tuple[PD, T], Tuple[PD, T]]
):
    """This class filters the persistence diagrams of a dataset by their homology dimension.

    Here we assume that the dataset is a tuple of (persistence diagram, label) and that
    the points in the diagram are sorted by ascending lifetime. This is an invariant of
    the OneHotEncodedPersistenceDiagram class but could go wrong if the diagrams are modified
    in a way that breaks this invariant.

    Args:
        homology_dimensions_to_filter:
            The homology dimensions of the points in the diagram that should be kept.
    """

    is_fitted: bool
    homology_dimensions_to_filter_by: List[int]

    def __init__(self, homology_dimensions_to_filter: List[int]):
        self.homology_dimensions_to_filter_by = homology_dimensions_to_filter
        self.is_fitted = False

    def fit_to_dataset(self, dataset: Dataset[Tuple[PD, T]]) -> None:
        """This method does nothing."""
        self.is_fitted = True

    def __call__(self, item: Tuple[PD, T]) -> Tuple[PD, T]:
        """Filters the persistence diagram by homology dimension.

        Args:
            item:
                A tuple of (persistence diagram, label).

        Returns:
            A tuple of (persistence diagram, label).
        """
        if not self.is_fitted:
            raise RuntimeError(
                "The filter is not fitted to any dataset. "
                "Please call fit_to_dataset() first."
            )
        out: List[Tensor] = []
        for homology_dimension in self.homology_dimensions_to_filter_by:
            out.append(
                item[0]
                .get_all_points_in_homology_dimension(homology_dimension)
                .get_raw_data()
            )

        return PD(torch.cat(out)), item[1]



from abc import ABC, abstractmethod
import json
import os
from typing import Generic, TypeVar
import warnings
import jsonpickle

from torch.utils.data import Dataset

from gdeep.data.transforming_dataset import TransformingDataset

R = TypeVar("R")
S = TypeVar("S")


class AbstractPreprocessing(ABC, Generic[R, S]):
    @abstractmethod
    def fit_to_dataset(self, dataset: Dataset[R]) -> None:
        pass

    @abstractmethod
    def __call__(self, x: R) -> S:
        pass

    def transform(self, x: R) -> S:
        return self(x)

    def attach_transform_to_dataset(self, dataset: Dataset[R]) -> Dataset[S]:
        return TransformingDataset(dataset, self.transform)

    def save_pretrained(self, path: str) -> None:
        with open(
            os.path.join(path, self.__class__.__name__ + ".json"), "w"
        ) as outfile:
            whole_class = jsonpickle.encode(self)  # type: ignore
            json.dump(whole_class, outfile)

    def load_pretrained(self, path: str) -> None:
        try:
            with open(
                os.path.join(path, self.__class__.__name__ + ".json"), "r"
            ) as infile:
                whole_class = json.load(infile)
                self = jsonpickle.decode(whole_class)  # type: ignore
        except FileNotFoundError:
            warnings.warn(
                "The transformation file does not exist; attempting to run"
                " the transformation anyway..."
            )



from typing import Dict, Any

from captum import attr

# Define the dataset factory class for the tori dataset and torchvision datasets
# using the factory design pattern
# https://realpython.com/factory-method-python/


class AttributionFactory(object):
    """Attribution factory class for the Captum integration.
    This factory will contain the attributions techniques of captum.
    """

    _builders: Dict[str, Any] = {}

    def __init__(self):
        pass

    def register_builder(self, key: str, builder: Any):
        """this method adds to the internal builders dictionary
        new dataloader builders
        """
        self._builders[key] = builder  # type: ignore

    def build(self, key: str, *args, **kwargs) -> Any:  # type: ignore
        """This method returns the DataLoader builder
        corresponding to the input key.

        Args:
            key:
                the name of the dataset
        """
        builder = self._builders.get(key)  # type: ignore
        if builder is None:
            raise ValueError(f"No builder registered for key {key}")
        return builder(*args, **kwargs)  # type: ignore


class AttributionBuilder(object):
    """Builder class for the torchvision dataset
    and all its variations"""

    def __init__(self, attr_name: str):
        self.attr_name = attr_name

    def __call__(self, *args, **kwargs):  # type: ignore
        return attr.__getattribute__(self.attr_name)(  # type: ignore
            *args, **kwargs  # type: ignore
        )


def get_attr(key: str, *args, **kwargs) -> Any:  # type: ignore
    """Get a dataset from the factory

    Args:
        key :
            The name of the dataset, corresponding
            to the key in the list of builders
        **kwargs:
            The keyword arguments to pass to the dataset builder

    Returns:
        captum.attr:
            The captum interpretability tool
    """
    factory = AttributionFactory()

    for attr_name in attr.__all__:
        factory.register_builder(attr_name, AttributionBuilder(attr_name))

    return factory.build(key, *args, **kwargs)  # type: ignore



import warnings
from typing import Any, Tuple, TypeVar

import torch
from gdeep.data.persistence_diagrams.one_hot_persistence_diagram import (
    OneHotEncodedPersistenceDiagram,
)
from torch.utils.data import Dataset

from ..abstract_preprocessing import AbstractPreprocessing
from ..transforming_dataset import TransformingDataset

from .normalization import _compute_mean_of_dataset

from gdeep.utility.custom_types import Tensor

PD = OneHotEncodedPersistenceDiagram

T = TypeVar("T")


class NormalizationPersistenceDiagram(
    AbstractPreprocessing[Tuple[PD, T], Tuple[PD, T]]
):
    """This class runs the standard normalisation on the birth and death coordinates
    of the persistence diagrams of a dataset accross all the homology dimensions.

    The one-hot encoded persistence diagrams are kept as is.
    """

    is_fitted: bool
    mean: Tensor
    stddev: Tensor
    num_homology_dimensions: int

    def __init__(self, num_homology_dimensions: int):
        self.is_fitted = False
        self.num_homology_dimensions = num_homology_dimensions

    def fit_to_dataset(self, dataset: Dataset[Tuple[PD, T]]) -> None:
        # compute the mean and set the last entries to zero
        self.mean = _compute_mean_of_dataset(
            TransformingDataset(
                dataset, lambda x: (x[0].get_raw_data().mean(dim=0), x[1])
            )
        )
        self.mean = self.mean * torch.tensor(
            [1.0, 1.0] + [0.0] * (self.num_homology_dimensions)
        )
        self.stddev = _compute_mean_of_dataset(
            TransformingDataset(
                dataset,
                lambda x: (((x[0].get_raw_data() - self.mean) ** 2).mean(dim=0), x[1]),
            )
        )
        self.stddev = self.stddev * torch.tensor(
            [1.0, 1.0] + [0.0] * (self.num_homology_dimensions)
        ) + torch.tensor([0.0, 0.0] + [1.0] * (self.num_homology_dimensions))
        self.stddev = torch.sqrt(self.stddev)
        self.is_fitted = True

    def __call__(self, item: Tuple[PD, T]) -> Tuple[PD, T]:
        if not self.is_fitted:
            raise RuntimeError(
                "The normalisation is not fitted to any dataset."
                " Please call fit_to_dataset() first."
            )
        if not torch.all(self.stddev > 0):
            warnings.warn("The standard deviation contains zeros! Adding 1e-7")
            self.stddev = self.stddev + 1e-7
        out: PD = PD(((item[0]._data - self.mean) / self.stddev).float())  # type: ignore
        return (out, item[1])



"""Testing for simpleNN."""
# License: GNU AGPLv3

import torch
from torch import nn

from gdeep.models import ModelExtractor
from ..simple_nn import FFNet
from gdeep.data.datasets import DatasetBuilder, DataLoaderBuilder


bd = DatasetBuilder(name="Blobs")
ds_tr, ds_val, _ = bd.build()
dl = DataLoaderBuilder([ds_tr, ds_val])  # type: ignore
dl_tr, dl_val, dl_ts = dl.build()  # sampler=SubsetRandomSampler(train_indices))


class Model1(nn.Module):
    def __init__(self):
        super(Model1, self).__init__()
        self.seqmodel = nn.Sequential(nn.Flatten(), FFNet(arch=(3, 5, 10, 5, 2)))

    def forward(self, x):
        return self.seqmodel(x)


def test_extractor_get_layers_param():
    model = Model1()

    loss_fn = nn.CrossEntropyLoss()
    me = ModelExtractor(model, loss_fn)

    lista = me.get_layers_param()

    # for k, item in lista.items():
    #     k, item.shape  # noqa


def test_extractor_get_decision_boundary():
    model = FFNet(arch=(3, 3))
    loss_fn = nn.CrossEntropyLoss()
    me = ModelExtractor(model, loss_fn)
    x = next(iter(dl_tr))[0][0]
    if x.dtype is not torch.int64:
        res = me.get_decision_boundary(x, n_epochs=5)
        assert res.shape[-1] == 3


def test_extractor_get_activations():
    model = FFNet(arch=(3, 4, 3))
    loss_fn = nn.CrossEntropyLoss()
    me = ModelExtractor(model, loss_fn)
    x = next(iter(dl_tr))[0][0]
    list_activations = me.get_activations(x)
    assert len(list_activations) == 3


def test_extractor_get_gradients():
    model = FFNet(arch=(3, 3, 4, 3))
    loss_fn = nn.CrossEntropyLoss()
    me = ModelExtractor(model, loss_fn)
    batch = next(iter(dl_tr))
    list_activations = me.get_gradients(batch)
    assert len(list_activations) == 2



import numpy as np
import sys
import torch

if sys.version_info >= (3, 10):
    from typing import TypeAlias  # type: ignore
else:
    from typing_extensions import TypeAlias  # type: ignore

Tensor: TypeAlias = torch.Tensor
Array: TypeAlias = np.ndarray



from IPython import get_ipython  # type: ignore
import base64
from typing import Dict, Optional, Type, List, Tuple, Union
import os
import time
import hashlib
import warnings

import torch
from torch import nn
from torch.utils.data import DataLoader
from torch.optim import Optimizer

from gdeep.utility.custom_types import Tensor


def _are_compatible(
    model_dict: Dict[str, Union[nn.Module, str]],
    dataloaders_dict: Dict[str, Union[List[DataLoader[Tuple[Tensor, Tensor]]], str]],
) -> bool:
    """utility function to check the compatibility of a model
    with a set of dataloaders `(dl_tr, dl_val, dl_ts)`
    """

    model = model_dict["model"]
    batch = next(iter(dataloaders_dict["dataloaders"][0]))[0]
    try:
        model(batch)  # type: ignore
    except RuntimeError:
        return False
    else:
        return True


def save_model_and_optimizer(
    model: nn.Module,
    model_name: Optional[str] = None,
    trial_id: Optional[str] = None,
    optimizer: Optional[Optimizer] = None,
    store_pickle: bool = False,
):
    """Save the model and the optimizer state_dict

    Args:
        model (nn.Module):
            the model to be saved
        model_name (str):
            model name
        trial_id (str):
            trial id to add to the name
        optimizer (torch.optim):
            the optimizer to save
        store_pickle (bool, default False):
            whether to store the pickle file of the model
            instead of the state_dict. The default
            is for state_dicts
    """

    if not trial_id:
        trial_id = str(round(time.time()))

    if os.path.exists("state_dicts"):
        pass
    else:
        os.makedirs("state_dicts")

    if model_name is None:
        if store_pickle:
            torch.save(
                model,
                os.path.join(
                    "state_dicts", model.__class__.__name__ + "-" + trial_id + ".pickle"
                ),
            )
        else:
            torch.save(
                model.state_dict(),
                os.path.join(
                    "state_dicts", model.__class__.__name__ + "-" + trial_id + ".pth"
                ),
            )
    else:
        if store_pickle:
            torch.save(
                model,
                os.path.join("state_dicts", model_name + "-" + trial_id + ".pickle"),
            )
        else:
            torch.save(
                model.state_dict(),
                os.path.join("state_dicts", model_name + "-" + trial_id + ".pth"),
            )
    if optimizer is not None:
        torch.save(
            optimizer.state_dict(),
            os.path.join(
                "state_dicts",
                str(optimizer)
                .replace("\n", "")
                .replace("(", "")
                .replace(":", "")
                .replace(")", "")
                + "-"
                + trial_id
                + ".pth",
            ),
        )


def ensemble_wrapper(clss: Type):
    """function to wrap the ensemble estimators
    of the ``torchensemble`` library.

    The only argument is the estimator class. Then
    you can initialise the output of this function
    as you would normally do for the original
    ``torchensemble`` class

    Args
        clss:
            the class of the estimator, like
            ``VotingClassifier`` for example
    Returns:
        type:
            the initialised ensemble estimator
            class that is compatible with giotto-deep
    """

    class NewEnsembleEstimator(clss):  # type: ignore
        def __init__(self, *args, **kwargs):
            super(NewEnsembleEstimator, self).__init__(*args, **kwargs)
            self.estimators_ = nn.ModuleList().extend(
                [self._make_estimator() for _ in range(self.n_estimators)]
            )

    return NewEnsembleEstimator


def _inner_refactor_scalars(list_, cross_validation, k_folds):
    """used to restructure lists of accuracies and losses
    per epoch"""
    out = []
    for t in range(len(list_)):
        lis = [x[0] for x in list_ if x[1] == t]
        value = sum(lis)
        if len(lis) > 0:
            if cross_validation:
                out.append([value / k_folds, t])
            else:
                out.append([value, t])
    return out


def is_notebook() -> bool:
    """Check if the current environment is a notebook

    Returns:
        bool:
            True if the environment is a notebook, False otherwise
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def autoreload_if_notebook() -> None:
    """Autoreload the modules if the environment is a notebook

    Returns:
        None
    """
    from IPython import get_ipython  # type: ignore

    get_ipython().magic("load_ext autoreload")  # type: ignore
    get_ipython().magic("autoreload 2")  # type: ignore


def _file_as_bytes(file) -> bytes:
    """Returns a bytes object representing the file

    Args:
        file (str):
            Path to the file

    Returns:
        bytes:
            Bytes object representing the file.
    """
    with open(file, "rb") as f:
        return f.read()


def get_checksum(file: str, encoding: str = "hex"):
    """Returns the checksum of the file

    Args:
        file :
            Path to the file
        encoding:
            string encoding like ``'hex'``

    Raises:
        ValueError: if the file does not exist
        ValueError: if the encoding is not supported

    Returns:
            The checksum of the file. If the file does not exist,
            None is returned.
    """
    # Check if file exists
    if not os.path.exists(file):
        raise ValueError("File {} does not exist".format(file))
    if encoding == "hex":
        return hashlib.md5(_file_as_bytes(file)).hexdigest()
    elif encoding == "base64":
        return base64.b64encode(
            bytes.fromhex(hashlib.md5(_file_as_bytes(file)).hexdigest())
        )
    else:
        raise ValueError("encoding must be either 'hex' or 'base64'")


class KnownWarningSilencer:
    """silence all warnings within this ``with``
    statement with this class"""

    def __init__(self) -> None:
        pass

    def __enter__(self) -> "KnownWarningSilencer":
        warnings.filterwarnings("ignore")
        return self

    def __exit__(self, type, value, traceback):
        warnings.filterwarnings("default")


def flatten_list_of_lists(list_: list) -> list:
    """Flatten a list of lists
    Args:
        list_ (list):
            the list to flatten
    Returns:
        list:
            the flattened list
    """
    return [item for sublist in list_ for item in sublist]



from typing import List, Tuple

from torch import nn
import torchvision.models as models
from torch.optim import SGD, Adam
from sklearn.model_selection import StratifiedKFold
from torch.utils.data.sampler import SubsetRandomSampler
from torch.utils.data import Dataset, DataLoader
import numpy as np
import torch

from gdeep.trainer import Trainer
from gdeep.data.datasets import DatasetBuilder, DataLoaderBuilder
from gdeep.data.preprocessors import ToTensorImage
from gdeep.search import HyperParameterOptimization, GiottoSummaryWriter
from gdeep.models import FFNet

from gdeep.utility.custom_types import Tensor


# parametric model with string value
class Model2(nn.Module):
    def __init__(self, n_nodes="100"):
        super(Model2, self).__init__()
        self.md = nn.Sequential(
            nn.Sequential(
                models.resnet18(pretrained=True), nn.Linear(1000, eval(n_nodes))
            ),
            nn.Linear(eval(n_nodes), 10),
        )

    def forward(self, x):
        return self.md(x)


# model
class Model1(nn.Module):
    def __init__(self):
        super(Model1, self).__init__()
        self.seqmodel = nn.Sequential(nn.Flatten(), FFNet(arch=(4, 5, 4)))

    def forward(self, x):
        return self.seqmodel(x).reshape(-1, 2, 2)


writer = GiottoSummaryWriter()

# download MNIST
bd = DatasetBuilder(name="CIFAR10")
ds_tr, _, _ = bd.build(download=True)

# Preprocessing
transformation = ToTensorImage([32, 32])
transformation.fit_to_dataset(ds_tr)  # this is useless for this transformation

transformed_ds_tr = transformation.attach_transform_to_dataset(ds_tr)

# use only 32*7 images from CIFAR10
train_indices = list(range(32 * 7))
dl_tr, *_ = DataLoaderBuilder([transformed_ds_tr]).build(  # type: ignore
    [{"batch_size": 32, "sampler": SubsetRandomSampler(train_indices)}]
)


def test_hpo_failure():
    # define the model
    model = Model2()

    # initialise loss
    loss_fn = nn.CrossEntropyLoss()

    # initialise pipeline class
    pipe = Trainer(
        model,
        [dl_tr, None],  # type: ignore
        loss_fn,
        writer,
        k_fold_class=StratifiedKFold(2, shuffle=True),
    )

    # initialise gridsearch
    try:
        HyperParameterOptimization(pipe, "accuracy", 2, best_not_last=True)
    except AssertionError:
        pass


def test_hpo_cross_val():
    # define the model
    model = Model2()

    # initialise loss
    loss_fn = nn.CrossEntropyLoss()

    # initialise pipeline class
    pipe = Trainer(
        model,
        [dl_tr, None],  # type: ignore
        loss_fn,
        writer,
        k_fold_class=StratifiedKFold(2, shuffle=True),
    )

    # initialise gridsearch
    search = HyperParameterOptimization(pipe, "accuracy", 2, best_not_last=True)

    # if you want to store pickle files of the models instead of the state_dicts
    search.store_pickle = True

    # dictionaries of hyperparameters
    optimizers_params = {"lr": [0.001, 0.01]}
    dataloaders_params = {"batch_size": [32, 64, 16]}
    models_hyperparams = {"n_nodes": ["200"]}

    # starting the gridsearch
    search.start(
        [SGD, Adam],
        3,
        False,
        optimizers_params,
        dataloaders_params,
        models_hyperparams,
        n_accumulated_grads=2,
    )


def test_hpo_accumulated_grads():
    # define the model
    model = Model2()

    # initialise loss
    loss_fn = nn.CrossEntropyLoss()

    # initialise pipeline class
    pipe = Trainer(model, [dl_tr, None], loss_fn, writer)  # type: ignore

    # initialise gridsearch
    search = HyperParameterOptimization(pipe, "accuracy", 2, best_not_last=True)

    # if you want to store pickle files of the models instead of the state_dicts
    search.store_pickle = True

    # dictionaries of hyperparameters
    optimizers_params = {"lr": [0.001, 0.01]}
    dataloaders_params = {"batch_size": [32, 64, 16]}
    models_hyperparams = {"n_nodes": ["200"]}

    # starting the gridsearch
    search.start(
        [SGD, Adam],
        2,
        False,
        optimizers_params,
        dataloaders_params,
        models_hyperparams,
        n_accumulated_grads=2,
    )


def test_hpo_loss():
    # define the model
    model = Model2()

    # initialise loss
    loss_fn = nn.CrossEntropyLoss()

    # initialise pipeline class
    pipe = Trainer(model, [dl_tr, None], loss_fn, writer)  # type: ignore

    # initialise gridsearch
    search = HyperParameterOptimization(pipe, "loss", 2, best_not_last=True)

    # dictionaries of hyperparameters
    optimizers_params = {"lr": [0.001, 0.01]}
    dataloaders_params = {"batch_size": [32, 64, 16]}
    models_hyperparams = {"n_nodes": ["200"]}

    # starting the gridsearch
    search.start(
        [SGD, Adam], 2, False, optimizers_params, dataloaders_params, models_hyperparams
    )


def test_hpo_string_parameters():
    # define the model
    model = Model2()

    # initialise loss
    loss_fn = nn.CrossEntropyLoss()

    # initialise pipeline class
    pipe = Trainer(model, [dl_tr, None], loss_fn, writer)  # type: ignore

    # initialise gridsearch
    search = HyperParameterOptimization(pipe, "loss", 2)

    # if you want to store pickle files of the models instead of the state_dicts
    search.store_pickle = True

    # dictionaries of hyperparameters
    optimizers_params = {"lr": [0.001, 0.01]}
    dataloaders_params = {"batch_size": [32, 64]}
    models_hyperparams = {"n_nodes": ["200", "256"]}

    # starting the gridsearch
    search.start(
        [SGD], 1, False, optimizers_params, dataloaders_params, models_hyperparams
    )


def test_hpo_collate():
    class MyDataset(Dataset):
        def __init__(self):
            self.x = []
            for _ in range(100):
                self.x.append((torch.rand(1, np.random.randint(2, 4))).to(torch.float))
            self.y = np.array(
                np.random.randint(2, size=100 * 2).reshape(-1, 2), dtype=np.int64
            )

        def __len__(self):
            return 100

        def __getitem__(self, item: int) -> Tuple[Tensor, Tensor]:
            return self.x[item], self.y[item]

    def collate_fn(batch_tuple: List):
        # print(batch_tuple)
        target = torch.zeros(len(batch_tuple), 4)
        label = torch.zeros(len(batch_tuple), 2).to(torch.long)
        for i, batch in enumerate(batch_tuple):
            source = batch[0]
            target[i, : len(source[-1])] = source
            label[i] = torch.tensor(batch[1]).to(torch.long)
        return target, label

    model = Model1()
    # dataloaders
    ds = MyDataset()
    dl_train = DataLoader(ds, batch_size=6, collate_fn=collate_fn)

    # loss function
    loss_fn = nn.CrossEntropyLoss()
    # pipeline
    pipe = Trainer(model, [dl_train, None], loss_fn, writer)  # type: ignore
    # initialise gridsearch
    search = HyperParameterOptimization(pipe, "loss", 2, best_not_last=True)

    # dictionaries of hyperparameters
    optimizers_params = {"lr": [0.001, 0.01]}
    dataloaders_params = {"batch_size": [3, 9, 2], "collate_fn": [collate_fn]}

    # starting the gridsearch
    search.start([SGD, Adam], 1, False, optimizers_params, dataloaders_params)



import os
from os import remove, listdir
from os.path import join
from shutil import rmtree

from gdeep.utility.constants import ROOT_DIR


def clean_up_files(fun):
    """Decorator to remove all the files created by the method it decorates"""

    def wrapper():
        # Get all subdirectories of the root directory
        directories = [d for d in listdir(ROOT_DIR) if os.path.isdir(join(ROOT_DIR, d))]

        # Get all files in the root directory
        files = [f for f in listdir(ROOT_DIR) if os.path.isfile(join(ROOT_DIR, f))]

        fun()
        # Get all subdirectories of the root directory
        new_directories = [
            d for d in listdir(ROOT_DIR) if os.path.isdir(join(ROOT_DIR, d))
        ]

        # Get all files in the root directory
        new_files = [f for f in listdir(ROOT_DIR) if os.path.isfile(join(ROOT_DIR, f))]

        # Delete all newly created directories
        for directory in new_directories:
            if directory not in directories:
                rmtree(join(ROOT_DIR, directory))

        # Delete all newly created files
        for file in new_files:
            if file not in files:
                remove(join(ROOT_DIR, file))

    return wrapper



"""Testing for simpleNN."""
# License: GNU AGPLv3
import pytest
import torch
from .. import FFNet


def test_FFNet():
    X = torch.rand(10, 2)
    sn = FFNet((2, 3, 2))
    out = sn.forward(X)
    assert out.shape[0] == 10 and out.shape[1] == 2



from .transforming_dataset import TransformingDataset
from .preprocessing_pipeline import PreprocessingPipeline
from .dataset_factory import DatasetFactory
from .abstract_preprocessing import AbstractPreprocessing
from ._utils import MissingVocabularyError


__all__ = [
    "TransformingDataset",
    "PreprocessingPipeline",
    "DatasetFactory",
    "AbstractPreprocessing",
    "MissingVocabularyError",
]



from abc import ABC, abstractmethod
from typing import Callable, Dict, Optional
from transformers.configuration_utils import PretrainedConfig
import torch
import torch.nn as nn
from torch.nn import (
    Module,
    MultiheadAttention,
    Linear,
    Sequential,
    LayerNorm,
    Dropout,
    ModuleList,
)

from gdeep.utility.enum_types import (
    ActivationFunction,
    PoolerType,
    AttentionType,
    LayerNormStyle,
)

# Type aliases
from gdeep.utility.custom_types import Tensor


class PersformerConfig(PretrainedConfig):
    """
    Configuration class to define a persformer model.

    Examples::

        from gdeep.topological_layers import PersformerConfig, PersformerModel
        # Initialize the configuration object
        config = PersformerConfig()
        # Initialize the model
        model = Persformer(config)
        # Access the configuration object
        config = model.config

    """

    input_size: int  # input size of the model
    output_size: int
    hidden_size: int
    num_attention_layers: int
    num_attention_heads: int
    intermediate_size: int
    hidden_act: ActivationFunction
    hidden_dropout_prob: float
    attention_probs_dropout_prob: float
    layer_norm_eps: float
    classifier_dropout_prob: float
    layer_norm_style: LayerNormStyle
    attention_type: AttentionType
    activation_fn: ActivationFunction
    pooler_type: PoolerType
    use_attention_only: bool
    use_skip_connections_for_persformer_blocks: bool

    def __init__(
        self,
        input_size: int = 2 + 4,
        output_size: int = 2,
        hidden_size: int = 32,
        num_attention_layers: int = 2,
        num_attention_heads: int = 4,
        intermediate_size: int = 32,
        hidden_act: ActivationFunction = ActivationFunction.GELU,
        hidden_dropout_prob: float = 0.1,
        attention_probs_dropout_prob: float = 0.1,
        layer_norm_eps: float = 1e-12,
        classifier_dropout_prob: float = 0.1,
        use_layer_norm: LayerNormStyle = LayerNormStyle.NO_LAYER_NORMALIZATION,
        attention_type: AttentionType = AttentionType.DOT_PRODUCT,
        pooler_type: PoolerType = PoolerType.ATTENTION,
        use_attention_only: bool = False,
        use_skip_connections_for_persformer_blocks=False,
        **kwargs
    ):
        super().__init__(**kwargs)  # type: ignore
        self.input_size = input_size
        self.output_size = output_size
        self.hidden_size = hidden_size
        self.num_attention_layers = num_attention_layers
        self.num_attention_heads = num_attention_heads
        self.intermediate_size = intermediate_size
        self.hidden_act = hidden_act
        self.hidden_dropout_prob = hidden_dropout_prob
        self.attention_probs_dropout_prob = attention_probs_dropout_prob
        self.layer_norm_eps = layer_norm_eps
        self.classifier_dropout_prob = classifier_dropout_prob
        self.layer_norm_style = use_layer_norm
        self.attention_type = attention_type
        self.pooler_type = pooler_type
        self.use_attention_only = use_attention_only
        self.use_skip_connections_for_persformer_blocks = (
            use_skip_connections_for_persformer_blocks
        )



import numpy as np
import torch
from gtda.homology import VietorisRipsPersistence

from .. import (
    OneHotEncodedPersistenceDiagram,
    get_one_hot_encoded_persistence_diagram_from_gtda,
)


def test_one_hot_encoded_persistence_diagram():
    pd_one_hot = OneHotEncodedPersistenceDiagram(
        torch.tensor(
            [
                [0.3, 0.5, 1.0, 0.0],
                [0.4, 0.8, 1.0, 0.0],
                [0.5, 0.9, 0.0, 1.0],
            ]
        )
    )

    lifetimes = pd_one_hot.get_lifetimes()

    assert torch.allclose(lifetimes, torch.tensor([0.2000, 0.4000, 0.4000]))

    pd_filtered = pd_one_hot.filter_by_lifetime(min_lifetime=0.3, max_lifetime=0.5)

    assert pd_filtered.all_close(
        OneHotEncodedPersistenceDiagram(
            torch.tensor(
                [[0.5000, 0.9000, 0.0000, 1.0000], [0.4000, 0.8000, 1.0000, 0.0000]]
            )
        )
    )


def test_from_numpy():
    x = np.array(
        [
            [0.3, 0.5, 1.0, 0.0],
            [0.4, 0.8, 1.0, 0.0],
            [0.5, 0.9, 0.0, 1.0],
        ]
    )

    x = OneHotEncodedPersistenceDiagram.from_numpy(x)

    assert x.all_close(
        OneHotEncodedPersistenceDiagram(
            torch.tensor(
                [
                    [0.3000, 0.5000, 1.0000, 0.0000],
                    [0.5000, 0.9000, 0.0000, 1.0000],
                    [0.4000, 0.8000, 1.0000, 0.0000],
                ]
            )
        )
    )


def test_get_one_hot_encoded_persistence_diagram_from_gtda():
    vr = VietorisRipsPersistence(homology_dimensions=[0, 1])

    # set ranom seed
    np.random.seed(0)

    points = np.random.rand(100, 2)  # type: ignore

    diagrams = vr.fit_transform([points])

    diagram = diagrams[0]

    assert get_one_hot_encoded_persistence_diagram_from_gtda(diagram)._data.shape == (
        123,
        4,
    )



from functools import reduce
from itertools import chain, combinations
import multiprocessing
from typing import Iterator, Any, Callable, Tuple, List, Optional, Union, Set

import torch
from torch import optim
from gph.python import ripser_parallel
from gtda.homology import FlagserPersistence as Flp
import plotly.figure_factory as ff
import plotly.graph_objects as go
from plotly.graph_objects import Figure
import numpy as np
from tqdm import tqdm
import operator as op

from gdeep.utility import DEVICE

from gdeep.utility.custom_types import Tensor

from gdeep.utility.custom_types import Array


class PersistenceGradient:
    """This class computes the gradient of the persistence
    diagram with respect to a point cloud. The algorithms has
    first been developed in https://arxiv.org/abs/2010.08356 .

    Disclaimer: this algorithm works well for generic point clouds.
    In case your point cloud has many simplices with same
    filtration values, the matching of the points to the persistent
    features may fail to disambiguate.

    Args:
        zeta :
            the relative weight of the regularisation part
            of the ``persistence_function``
        homology_dimensions :
            tuple of homology dimensions
        collapse_edges :
            whether to use Collapse or not. Not implemented yet.
        max_edge_length :
            the maximum edge length
            to be consider not infinity
        approx_digits :
            digits after which to trunc floats for
            list comparison
        metric :
            either ``"euclidean"`` or ``"precomputed"``.
            The second one is in case of ``x`` being
            the pairwise-distance matrix or
            the adjacency matrix of a graph.
        directed :
            whether the input graph is a directed graph
            or not. Relevant only if ``metric = "precomputed"``

    Examples::

        from gdeep.utility.optimization import PersistenceGradient
        # prepare the datum
        X = torch.tensor([[1, 0.], [0, 1.], [2, 2], [2, 1]])
        # select the homology dimensions
        hom_dim = [0, 1]
        # initialise the class
        pg = PersistenceGradient(homology_dimensions=hom_dim,
                                 zeta=0.1,
                                 max_edge_length=3,
                                 collapse_edges=False)
        # run the optimisation for four epochs!
        pg.sgd(X, n_epochs=4, lr=0.4)

    """

    dist_mat: Tensor

    def __init__(
        self,
        homology_dimensions: Optional[List[int]],
        zeta: float = 0.5,
        collapse_edges: bool = False,
        max_edge_length: float = np.inf,
        approx_digits: int = 6,
        metric: str = "euclidean",
        directed: bool = False,
    ) -> None:

        self.collapse_edges = collapse_edges
        self.max_edge_length = max_edge_length
        self.metric = metric
        self.directed = directed
        self.approx_digits = approx_digits
        self.zeta = zeta
        if homology_dimensions:
            self.homology_dimensions = homology_dimensions
        else:
            self.homology_dimensions = [0, 1]

    @staticmethod
    def powerset(iterable: Iterator, max_length: int) -> Iterator:
        """powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)
        up to `max_length`."""
        s = list(iterable)
        return chain.from_iterable(combinations(s, r) for r in range(0, max_length + 1))

    @staticmethod
    def _parallel_apply_along_axis(
        func1d: Callable, axis: int, arr: Array, *args: Any, **kwargs: Any
    ) -> Array:
        """
        Like ``numpy.apply_along_axis()``, but takes advantage of multiple
        cores.
        """
        # Effective axis where apply_along_axis() will be applied by each
        # worker (any non-zero axis number would work, so as to allow the use
        # of `np.array_split()`, which is only done on axis 0):
        effective_axis = 1 if axis == 0 else axis
        if effective_axis != axis:
            arr = arr.swapaxes(axis, effective_axis)
        n_processes = min(len(arr) // 2, multiprocessing.cpu_count())
        # Chunks for the mapping (only a few chunks):
        chunks = [
            (func1d, effective_axis, sub_arr, args, kwargs)
            for sub_arr in np.array_split(arr, n_processes)
        ]
        pool = multiprocessing.Pool(n_processes)
        individual_results = pool.map(unpacking_apply_along_axis, chunks)  # type: ignore
        # Freeing the workers:
        pool.close()
        pool.join()
        return np.concatenate(individual_results)  # type: ignore

    def _simplicial_pairs_of_indices(self, x: Tensor) -> Tensor:
        """Private function to compute the pair of indices in X to
        matching the simplices.
        """
        simplices = list(
            self.powerset(
                list(range(0, len(x))),  # type: ignore
                max(self.homology_dimensions) + 2,
            )
        )[1:]
        simplices_array = np.array(simplices, dtype=object).reshape(-1, 1)
        comb_number = comb(max(self.homology_dimensions) + 2, 2)
        # the current computation bottleneck
        if len(simplices_array) > 10000000:
            pairs_of_indices = self._parallel_apply_along_axis(
                _combinations_with_single, 1, simplices_array, comb_number
            )
        else:
            pairs_of_indices = np.apply_along_axis(
                _combinations_with_single, 1, simplices_array, comb_number
            )
        return torch.tensor(pairs_of_indices, dtype=torch.int64)

    def phi(self, x: Tensor) -> Tensor:
        """This function is from :math:`(R^d)^n` to :math:`R^{|K|}`,
        where K is the top simplicial complex of the VR filtration.
        It is defined as:
        :math:`\\Phi_{\\sigma}(X)=max_{i,j \\in \\sigma}||x_i-x_j||.`

        Args:
            x:
                the argument of :math:`\\Phi`, a tensor

        Returns:
            Tensor:
                the value of :math:`\\Phi` at ``x``
        """

        if self.metric == "precomputed":
            self.dist_mat = x
        else:
            self.dist_mat = torch.cdist(x, x)
        simplicial_pairs = self._simplicial_pairs_of_indices(x).reshape(-1, 2)
        ks = simplicial_pairs[:, 0]
        js = simplicial_pairs[:, 1]
        comb_number = comb(max(self.homology_dimensions) + 2, 2)
        # morally, this is what one would like to do:
        # lista = [max([dist_mat[pair] for pair in pairs]) for pairs in
        #    simplicial_pairs if max([dist_mat[pair] for pair in pairs])
        #    <= self.max_edge_length]
        lista = torch.max(
            (
                torch.gather(
                    torch.index_select(self.dist_mat, 0, ks), 1, js.reshape(-1, 1)
                )
            ).reshape(-1, comb_number),
            1,
        )[0]
        lista = torch.sort(lista)[0]
        return lista

    @staticmethod
    def _compute_pairs(x: Tensor) -> Tuple[Tensor, Tensor]:
        """Use giotto-tda to compute homology (b,d) pairs

        Args:
            x:
                this is the input point cloud or the input
                pairwise distance or the adjacency matrix of a graph

        Returns:
            Tensor, Tensor:
                the first tensor is the tensor obtained from
                the list of pairs ``(b,d)``; the second 1d tensor
                contains the homology
                group index"""

        dgms = Flp().fit_transform([x.detach().numpy()])
        pairs = dgms[0]
        return pairs[:, :2], pairs[:, 2]

    def _persistence(self, x: Tensor) -> Tuple[Array, Tensor, Tensor]:
        """This function computes the persistence permutation.
        This permutation permutes the filtration values and matches
        them as follows:
        :math:`Pers:Filt_K \\subset \\mathbb R^{|K|} \\to (\\mathbb R^2)^p
        \\times \\mathbb R^q, \\Phi(X) \\mapsto D = \\cup_i^p
        (\\Phi_{\\sigma_i1}(X) ,\\Phi_{\\sigma_i2}(X) )
        \\times \\cup_j^q (\\Phi_{\\sigma_j}(X),+\\infty).`

        Args:
            x:
                this is the point cloud

        Returns:
            persistence_pairs :
                this is an array of pairs (i,j). ``i``
                and ``j`` are the indices in the list ``phi`` associated to a
                positive-negative pair. When ``j`` is equal to ``-1``, it means
                that the match was not found and the feature
                is infinitely persistent. When ``(i,j)=(-1,-1)``, simply
                discard the pair
            phi :
                this is the 1d tensor of all filtration values
            homology_dims :
                this 1d tensor contains the homology
                group index
        """
        phi = self.phi(x)
        pairs, homology_dims = self._compute_pairs(x)
        approx_pairs: List[float] = list(np.around(pairs, self.approx_digits).flatten())
        num_approx = 10**self.approx_digits
        inv_num_approx = 10 ** (-self.approx_digits)
        phi_approx: Tensor = torch.round(phi * num_approx) * inv_num_approx
        # find the indices of phi_approx elements that are in approx_pairs
        indices_in_phi_of_pairs_ls: List = []
        for i in approx_pairs:
            indices_in_phi_of_pairs_ls += (
                torch.nonzero(phi_approx == i).flatten().detach().tolist()
            )
        indices_in_phi_of_pairs: Set = set(indices_in_phi_of_pairs_ls)
        persistence_pairs_array = -np.ones((len(approx_pairs), 2), dtype=np.int32)
        for i in indices_in_phi_of_pairs:
            try:
                temp_index: int = approx_pairs.index(
                    round(phi[i].item(), self.approx_digits)  # type: ignore
                )
                approx_pairs[temp_index] = float("inf")
                persistence_pairs_array[temp_index // 2, temp_index % 2] = int(i)
            except ValueError:
                pass
        return persistence_pairs_array, phi, homology_dims

    def _computing_persistence_with_gph(self, xx: Tensor) -> List:
        """This method accepts the pointcloud and returns the
        persistence diagram in the following form
        :math:`Pers:Filt_K \\subset \\mathbb R^{|K|} \\to (\\mathbb R^2)^p
        \\times \\mathbb R^q, \\Phi(X) \\mapsto D = \\cup_i^p
        (\\Phi_{\\sigma_i1}(X) ,\\Phi_{\\sigma_i2}(X) )
        \\times \\cup_j^q (\\Phi_{\\sigma_j}(X),+\\infty).`
        The persistence diagram can be readily used for
        gradient descent.

        Args:
            xx:
                point cloud with ``shape = (n_points, n_features)``

        Returns:
            list of shape (n, 3):
                Persistence pairs (corresponding to the
                first 2 dimensions) where the last dimension
                contains the homology dimension
        """
        output = ripser_parallel(
            xx.detach().numpy(),
            maxdim=max(self.homology_dimensions),
            thresh=self.max_edge_length,
            coeff=2,
            metric=self.metric,
            collapse_edges=self.collapse_edges,
            n_threads=-1,
            return_generators=True,
        )

        persistence_pairs = []

        for dim in self.homology_dimensions:
            if dim == 0:
                persistence_pairs += [
                    (0, torch.norm(xx[x[1]] - xx[x[2]]), 0) for x in output["gens"][dim]
                ]
            else:
                persistence_pairs += [
                    (
                        torch.norm(xx[x[1]] - xx[x[0]]),
                        torch.norm(xx[x[3]] - xx[x[2]]),
                        dim,
                    )
                    for x in output["gens"][1][dim - 1]
                ]
        return persistence_pairs

    def persistence_function(self, xx: Tensor) -> Tensor:
        """This is the Loss function to optimise.
        :math:`L=-\\sum_i^p |\\epsilon_{i2}-\\epsilon_{i1}|+
        \\lambda \\sum_{x \\in X} ||x||_2^2`
        It is composed of a regularisation term and a
        function on the filtration values that is (p,q)-permutation
        invariant.

        Args:
            xx:
                this is the persistence function argument, a tensor

        Returns:
            Tensor:
                the function value at ``xx``

        """
        out: Tensor = torch.tensor(0).to(torch.float)
        persistence_array: Union[List, Array]
        # this is much slower
        if self.directed and self.metric == "precomputed":
            persistence_array, phi, _ = self._persistence(xx)
            for item in persistence_array:  # type: ignore
                if item[1] != -1:
                    out += phi[item[1]] - phi[item[0]]
        else:
            persistence_array = self._computing_persistence_with_gph(xx)
            for item in persistence_array:
                out += item[1] - item[0]
        reg = (xx**2).sum()  # regularisation term
        return -out + self.zeta * reg  # maximise persistence

    def sgd(
        self, xx: Tensor, lr: float = 0.01, n_epochs: int = 5
    ) -> Tuple[Figure, Figure, List[float]]:  # type: ignore
        """This function is the core function of this class and uses the
        SGD method to move points around in order to optimise
        ``persistence_function``

        Args:
            xx:
                2d tensor representing the point cloud,
                the first dimension is ``n_points`` while the second
                ``n_features``
            lr:
                learning rate for the SGD
            n_epochs:
                the number of gradient iterations

        Returns:
           fig, fig3d, loss_val:
               respectively the plotly ``quiver_plot``, plotly ``cone_plot``
               ad the list of loss function values over the epochs"""

        if not type(xx) == Tensor:
            xx = torch.tensor(xx)
        xx.to(DEVICE)
        xx.requires_grad = True
        x: Array = np.array([])
        z: Array = np.array([])
        y: Array = np.array([])
        u: Array = np.array([])
        v: Array = np.array([])
        w: Array = np.array([])
        loss_val = []
        optimizer = optim.Adam([xx], lr=lr)
        for _ in tqdm(range(n_epochs)):
            optimizer.zero_grad()
            loss = self.persistence_function(xx)
            loss_val.append(loss.item())
            x = np.concatenate((x, xx.detach().numpy()[:, 0]))  # type: ignore
            y = np.concatenate((y, xx.detach().numpy()[:, 1]))  # type: ignore
            loss.backward()  # compute gradients and store them in Xp.grad
            grads = -xx.grad.detach()  # type: ignore
            u = np.concatenate((u, 1 / grads.norm(2, 1).mean() * grads.numpy()[:, 0]))  # type: ignore
            v = np.concatenate((v, 1 / grads.norm(2, 1).mean() * grads.numpy()[:, 1]))  # type: ignore
            try:
                z = np.concatenate((z, xx.detach().numpy()[:, 2]))  # type: ignore
                w = np.concatenate((w, 1 / grads.norm(2) * grads.numpy()[:, 2]))  # type: ignore
            except IndexError:
                z = np.concatenate((z, 0 * xx.detach().numpy()[:, 1]))  # type: ignore
                w = np.concatenate((w, 0 * grads.numpy()[:, 1]))  # type: ignore
            optimizer.step()
        fig = ff.create_quiver(x, y, u, v)
        fig3d = Figure(
            data=go.Cone(
                x=x,
                y=y,
                z=z,
                u=u,
                v=v,
                w=w,
                sizemode="absolute",
                sizeref=2,
                anchor="tip",
            )
        )
        return fig, fig3d, loss_val


# this function is outside the main class to use multiprocessing
def unpacking_apply_along_axis(tupl: List[Any]) -> Array:
    """
    Like numpy.apply_along_axis(), but with arguments in a tuple
    instead.

    This function is useful with multiprocessing.Pool().map(): (1)
    map() only handles functions that take a single argument, and (2)
    this function can generally be imported from a module, as required
    by map().
    """
    func1d, axis, arr, args, kwargs = tupl
    return np.apply_along_axis(func1d, axis, arr, *args, **kwargs)


def _combinations_with_single(tupl, max_length: int) -> Array:
    """Private function to compute combinations also for singletons
    with repetition. The un-intuitive shape of the output is
    becaue of the padding and vectorized operation happening
    in the next steps.

    Args:
        tupl :
            this is a list with only one element and such
            element is a tuple
        max_length :
            this is the paddding length.

    Returns:
        ndarray :
            padded list -- with (0,0) -- of all the combinations of indices
            within a given simplex."""
    tupl = tupl[0]  # the slice dimension will always contain 1 element
    if len(tupl) == 1:
        list1 = list(((tupl[0], tupl[0]),)) + list(
            (max_length - 1) * [(tupl[0], tupl[0])]
        )
        return np.array(list1)
    temp_list = list(combinations(tupl, 2))
    list2 = temp_list + ((max_length - len(temp_list)) * [(tupl[0], tupl[1])])
    return np.array(list2)


def comb(nnn: int, kkk: int) -> int:
    """this function computes :math:`\\frac{n!}{(k!*(n-k)!)}` efficiently"""
    kkk = min(kkk, nnn - kkk)
    numer = reduce(op.mul, range(nnn, nnn - kkk, -1), 1)
    denom = reduce(op.mul, range(1, kkk + 1), 1)
    return numer // denom



import os
from typing import Tuple

import torch.nn as nn
from gdeep.data import PreprocessingPipeline
from gdeep.data.datasets import PersistenceDiagramFromFiles
from gdeep.data.datasets.base_dataloaders import (
    DataLoaderBuilder,
    DataLoaderParamsTuples,
)
from gdeep.data.datasets.persistence_diagrams_from_graphs_builder import (
    PersistenceDiagramFromGraphBuilder,
)
from gdeep.data.persistence_diagrams import (
    OneHotEncodedPersistenceDiagram,
    collate_fn_persistence_diagrams,
)
from gdeep.data.preprocessors import (
    FilterPersistenceDiagramByHomologyDimension,
    FilterPersistenceDiagramByLifetime,
    NormalizationPersistenceDiagram,
)
from gdeep.topology_layers import Persformer, PersformerConfig
from gdeep.trainer.trainer import Trainer
from gdeep.utility import DEFAULT_GRAPH_DIR, PoolerType
from sklearn.model_selection import train_test_split
from torch.optim import Adam
from torch.utils.data import Subset
from torch.utils.tensorboard.writer import SummaryWriter

from ..persformer import Persformer


def test_persformer_training():
    # Parameters
    name_graph_dataset: str = "MUTAG"
    diffusion_parameter: float = 0.1
    num_homology_types: int = 4

    # Create the persistence diagram dataset
    pd_creator = PersistenceDiagramFromGraphBuilder(
        name_graph_dataset, diffusion_parameter
    )
    pd_creator.create()

    pd_mutag_ds = PersistenceDiagramFromFiles(
        os.path.join(
            DEFAULT_GRAPH_DIR, f"MUTAG_{diffusion_parameter}_extended_persistence"
        )
    )

    # Create the train/validation/test split
    train_indices, test_indices = train_test_split(
        range(len(pd_mutag_ds)),
        test_size=0.2,
        random_state=42,
    )

    train_indices, validation_indices = train_test_split(
        train_indices,
        test_size=0.2,
        random_state=42,
    )

    # Create the data loaders
    train_dataset = Subset(pd_mutag_ds, train_indices)
    validation_dataset = Subset(pd_mutag_ds, validation_indices)
    test_dataset = Subset(pd_mutag_ds, test_indices)

    # Preprocess the data
    preprocessing_pipeline = PreprocessingPipeline[
        Tuple[OneHotEncodedPersistenceDiagram, int]
    ](
        (
            FilterPersistenceDiagramByHomologyDimension[int]([0, 1]),
            FilterPersistenceDiagramByLifetime[int](
                min_lifetime=-0.1, max_lifetime=1.0
            ),
            NormalizationPersistenceDiagram[int](num_homology_dimensions=4),
        )
    )

    preprocessing_pipeline.fit_to_dataset(train_dataset)

    train_dataset = preprocessing_pipeline.attach_transform_to_dataset(train_dataset)
    validation_dataset = preprocessing_pipeline.attach_transform_to_dataset(
        validation_dataset
    )
    test_dataset = preprocessing_pipeline.attach_transform_to_dataset(test_dataset)

    dl_params = DataLoaderParamsTuples.default(
        batch_size=32,
        num_workers=0,
        collate_fn=collate_fn_persistence_diagrams,
        with_validation=True,
    )

    # Build the data loaders
    dlb = DataLoaderBuilder((train_dataset, validation_dataset, test_dataset))  # type: ignore
    dl_train, dl_val, dl_test = dlb.build(dl_params)  # type: ignore

    # Define the model
    model_config = PersformerConfig(
        num_layers=6,
        num_attention_heads=4,
        input_size=2 + num_homology_types,
        ouptut_size=2,
        pooler_type=PoolerType.SUM,
    )

    model = Persformer(model_config)
    writer = SummaryWriter()

    loss_function = nn.CrossEntropyLoss()

    trainer = Trainer(model, [dl_train, dl_val, dl_test], loss_function, writer)

    trainer.train(
        Adam,
        3,
        False,
        {"lr": 0.01},
        {"batch_size": 16, "collate_fn": collate_fn_persistence_diagrams},
    )



from .utility import UniformlySampledPoint
from .decision_boundary_calculator import (
    DecisionBoundaryCalculator,
    GradientFlowDecisionBoundaryCalculator,
    QuasihyperbolicDecisionBoundaryCalculator,
)

# from .compute_topology import


__all__ = [
    "UniformlySampledPoint",
    "DecisionBoundaryCalculator",
    "GradientFlowDecisionBoundaryCalculator",
    "QuasihyperbolicDecisionBoundaryCalculator",
]



from typing import Any, Callable, Generic, TypeVar

from torch.utils.data import Dataset

R = TypeVar("R")
S = TypeVar("S")
T = TypeVar("T")


class TransformingDataset(Dataset[S], Generic[R, S]):
    """This class is the base class for the all
    Datasets that need to be transformed via preprocessors.
    This base class expects to get data from Dataset.

    Args:
        dataset :
            The source dataset for this class.
        transform :
            This is either a function defined by the
            users or a fitted preprocessor. The
            preprocessors inherits from ``AbstractPreprocessing``
    """

    dataset: Dataset[R]
    transform: Callable[[R], S]

    def __init__(self, dataset: Dataset[R], transform: Callable[[R], S]) -> None:
        self.dataset = dataset
        self.transform = transform  # type: ignore

    def __len__(self) -> int:
        if hasattr(self.dataset, "__len__"):
            return len(self.dataset)  # type: ignore
        else:
            raise NotImplementedError(
                "The dataset does not implement the __len__ method."
            )

    def __getitem__(self, idx: int) -> S:
        """The output of this method is one element
        in the dataset. The type of this element will
        change accordingly to the dataset itself. For
        example, in case of a text classification
        dataset, the output would probably be a tuple
        like ``(label, string)``
        """
        return self.transform(self.dataset[idx])  # type: ignore

    # forward all other methods of the TransformingDataset to the Dataset
    def __getattr__(self, name: str) -> Any:
        return getattr(self.dataset, name)


def append_transform(
    dataset: TransformingDataset[R, S], transform: Callable[[S], T]
) -> TransformingDataset[R, T]:
    """This function allows to concatenate different preprocessors
    that the usr may want to stack.

    Args:
        dataset :
            A dataset on which it is possible to apply transforms
        transform :
            A callable that can be applied to each item in the
            ``dataset``

    Returns:
        TransformingDataset:
            A new TransformingDataset that has been transformed
            by the transformation
    """

    def new_transform(x: R) -> T:
        return transform(dataset.transform(x))  # type: ignore

    return TransformingDataset(dataset.dataset, new_transform)
